import argparse
import logging
import torch
import time
import os
import os.path as osp
import sys
import yaml
import numpy as np
import pandas as pd
import itertools
from subprocess import run
from appdirs import user_data_dir
from tqdm import tqdm
from tempfile import NamedTemporaryFile
from copy import deepcopy
from urllib.request import urlretrieve
from .fusion_model import PL_fusion
from .rnn_model import PL_Sequence
from .preprocess import make_data_from_pdb, make_data_from_csv, make_annotations
from .structure_modeling import check_hlas, supported_hlas

DEVICE = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
DEFAULT_DIR = osp.dirname(osp.realpath(__file__))
MODEL_DIR = os.environ.get("SETTMHC_MODEL_DIR")
if MODEL_DIR is None:
    MODEL_DIR = user_data_dir("settmhc")
if not osp.exists(MODEL_DIR):
    os.mkdir(MODEL_DIR)
with open(osp.join(DEFAULT_DIR, 'hparams.yaml')) as f:
    CONFIG = yaml.load(f, Loader=yaml.FullLoader)


parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('--peptides', nargs='*', metavar='FSDVGEVPY',
                    help="Peptides for SETTMHC prediction, length from 8-11")
parser.add_argument('--hlas', nargs='*', metavar='HLA-A*02:01',
                    help="HLA alleles for SETTMHC prediction")
parser.add_argument('-i', '--input', metavar='input.csv',
                    help="Input file for prediction, see demo/demo.csv")
parser.add_argument('-o', '--output', metavar='output.csv',
                    help="Output file for results")
parser.add_argument('--sequence_only', action='store_true',
                    help="Use only sequence model to predict, but it's not recommended")
parser.add_argument('--download', action='store_true',
                    help="Download model files from Zenodo, default download to 'user_data_dir'")
parser.add_argument('--list_supported_hlas', action='store_true',
                    help="List all the supported 110 HLA alleles.")

pdb_args = parser.add_argument_group(
    title="Input single PDB file for prediction")
pdb_args.add_argument(
    '--pdb', help=" Single PDB/ent file for pMHC complex structure ('--peptides' and '--hlas' should also be given)")
pdb_args.add_argument('--protein', help='Protein chain in the input PDB file')
pdb_args.add_argument('--ligand', help='Ligand chain in the input PDB file')


def predict(cnn_data, rnn_data, skipped_idx):
    test_preds = np.zeros((5, len(cnn_data)))
    for i in range(1, 6):
        ckpt_path = osp.join(MODEL_DIR, f'model/Fusion/model{i}.ckpt')
        model = PL_fusion.load_from_checkpoint(
            checkpoint_path=ckpt_path, config=CONFIG, total_steps=8888, strict=False, map_location=DEVICE)
        model.eval()
        with torch.no_grad():
            pred = model(cnn_data, rnn_data)
            test_preds[i-1] = pred.squeeze().numpy()
    test_preds[:, skipped_idx] = 0
    preds = np.mean(test_preds, axis=0)
    return preds


def seq_predict(rnn_data):
    test_preds = np.zeros((5, len(rnn_data)))
    for i in range(1, 6):
        ckpt_path = osp.join(MODEL_DIR, f'model/RNN/model{i}.ckpt')
        model = PL_Sequence.load_from_checkpoint(
            checkpoint_path=ckpt_path, config=CONFIG, total_steps=8888, strict=False, map_location=DEVICE)
        model.eval()
        with torch.no_grad():
            pred = model(rnn_data)
            test_preds[i-1] = pred.squeeze().numpy()
    preds = np.mean(test_preds, axis=0)
    return preds


class DownloadBar(tqdm):
    def download_update(self, blocknum=1, blocksize=1, totalsize=None):
        if totalsize is not None:
            self.total = totalsize
        self.update(blocknum * blocksize - self.n)


def download_models():
    MODEL_URL = 'https://zenodo.org/record/7226854/files/model.tar.gz?download=1'
    temp = NamedTemporaryFile(delete=False, suffix='.tar.gz')
    print(f"Download model file in {temp.name}")

    urlretrieve(MODEL_URL, temp.name, reporthook=DownloadBar(
        unit='B', unit_scale=True).download_update)
    temp.close()
    run(f"tar -xf {temp.name} -C {MODEL_DIR}", shell=True)
    print(
        f'Extracted {temp.name} to {MODEL_DIR}, you may delete the temp file manually')


def check_models():
    return osp.exists(osp.join(MODEL_DIR, 'model'))


def main(argv=sys.argv[1:]):
    args = parser.parse_args(argv)
    if args.download:
        download_models()
        already_download = check_models()
        if not already_download:
            print(
                "Downloads for models failed. May visit https://zenodo.org/record/7226854 manually")
        return
    else:
        already_download = check_models()
        if not already_download:
            print(
                f"There is no file in {MODEL_DIR}. Must use --download to get models for the first application")
            return

    if args.list_supported_hlas:
        print(supported_hlas)
        return

    if args.pdb:
        if args.sequence_only:
            parser.error(
                "If a PDB file is specified, sequence_only must not be specified")
        if not args.peptides or not args.hlas:
            parser.error(
                "If a PDB file is specified, peptide and hla must be given")
        check_hlas(args.hlas)
        cnn_data, rnn_data = make_data_from_pdb(
            args.peptides[0], args.hlas[0], args.pdb, args.protein, args.ligand)
        cnn_data = cnn_data.unsqueeze(0)
        df = pd.DataFrame({'peptide': args.peptides, 'HLA': args.hlas})
    elif args.input:
        if args.peptides or args.hlas:
            parser.error(
                "If an input csv file is specified, peptides or hlas should not be specified")
        df = pd.read_csv(args.input)
        check_hlas(df['HLA'])
        annos = df['Annotation'] if 'Annotation' in df.columns else make_annotations(
            len(df))
        df['Annotation'] = annos
        cnn_data, rnn_data, skipped_idx = make_data_from_csv(
            df, sequence_only=args.sequence_only)
    else:
        if not args.peptides or not args.hlas:
            parser.error(
                "Either specify an input csv file or peptides and hlas")
        check_hlas(args.hlas)
        ls = list(itertools.product(args.peptides, args.hlas))
        df = pd.DataFrame(ls, columns=['peptide', 'HLA'])
        annos = df['Annotation'] if 'Annotation' in df.columns else make_annotations(
            len(df))
        df['Annotation'] = annos
        cnn_data, rnn_data, skipped_idx = make_data_from_csv(
            df, sequence_only=args.sequence_only)

    # make predictions
    if args.sequence_only:
        preds = seq_predict(rnn_data)
    else:
        preds = predict(cnn_data, rnn_data, skipped_idx)

    output_df = deepcopy(df)
    output_df['Prediction_scores'] = preds
    if not args.output:
        now = time.strftime("%Y%m%d-%H_%M", time.localtime(time.time()))
        output_file = f'{now}_predictions.csv'
        logging.warning(
            f'Output file is not specified, the results will be saved in {output_file}')
    else:
        output_file = args.output
    output_df.to_csv(output_file, index=False)



# pymol -qc mutate.py pdb selection new_residue
# example: pymol -qc mutate.py 0.pdb C/1/ ALA 0_mutated.pdb

# run for all the confs in the ensemble
# append all to one using mdconvert

from pymol import cmd
import sys


cmd.load(sys.argv[-2], "mobile")
cmd.load(sys.argv[-1], "ref")

cmd.align("mobile & (chain A | chain B)", "ref & (chain A | chain B)")

cmd.save("aln-"+str(sys.argv[-2]), "mobile")
#cmd.save("aln-"+str(sys.argv[4]), "end")

"""
pdb, selection, mutant, output_name = sys.argv[-4:]
print "file:", pdb
print "selection:", selection
print "mutating to:", mutant
print "output name:", output_name

cmd.load(pdb, "file")

cmd.wizard("mutagenesis")
cmd.refresh_wizard()
cmd.get_wizard().do_select(selection)
cmd.get_wizard().set_mode(mutant)
cmd.get_wizard().apply()
cmd.set_wizard()

cmd.save(output_name, "file")
"""


#! /home/snerli/anaconda3/bin/python
# additional bio libraries

from pyrosetta import *

#custom libraries
from scoring.rmsd import RMSD
from ia.chain_split import CHAIN_SPLIT
from ia.interface_analyzer import INTERFACE

# import other required libraries
import os
import sys
import fcntl
import argparse

def extract_residues_from_movemap(movemap):
    readfilehandler = open(movemap, "r")
    resi_list = pyrosetta.rosetta.std.list_unsigned_long_t()
    for line in readfilehandler:
        line = line.rstrip()
        if "RESIDUE" in line and "CHI" in line:
            resi = line.split(" ")[1]
            resi_list.append(int(resi))
    readfilehandler.close()
    return resi_list

def rmsd(my_pose, args):
    native_pose = pose_from_pdb(args.native)

    rms_value = 0
    rms_obj = RMSD(my_pose, native_pose)

    rms_all = rms_obj.all_atom_rmsd()
    rms_ca = rms_obj.CA_rmsd()
    resi_list = extract_residues_from_movemap(args.movemap)
    rms_ca_movemap = rms_obj.CA_rmsd_residue_number(resi_list)

    return [rms_all, rms_ca, rms_ca_movemap]

def ia(my_pose, args):
    chain = CHAIN_SPLIT(my_pose, args.cutpoint)
    chain.cut()
    split = chain.get_pose()

    ia = INTERFACE(split)
    ia.analyze()
    return ia.get_dG()

def write_to_file(args, tag, bind_energy, rms_all, rms_ca, rms_ca_movemap):
    with open(args.outfile, "a") as writefilehandler:
        fcntl.flock(writefilehandler, fcntl.LOCK_EX)
        writefilehandler.write(tag+","+str(bind_energy)+","+str(rms_all)+","+str(rms_ca)+","+str(rms_ca_movemap)+"\n")
        fcntl.flock(writefilehandler, fcntl.LOCK_UN)


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='compute rmsd values')
    parser.add_argument("-pdb", help="provide the pdb files")
    parser.add_argument("-native", help="provide native pdb file")
    parser.add_argument("-movemap", help="find rmsd value for the residues used in the movemap")
    parser.add_argument("-cutpoint", help="provide cutpoint", type=int)
    parser.add_argument("-outfile", help="provide output file name", default="rmsds_ia.csv")
    args = parser.parse_args()

    init()

    my_pose = pose_from_pdb(args.pdb)
    [rms_all, rms_ca, rms_ca_movemap] = rmsd(my_pose, args)
    bind_energy = ia(my_pose, args)
    tag = args.pdb.split(".pdb")[0]
    write_to_file(args, tag, bind_energy, rms_all, rms_ca, rms_ca_movemap)



#!/usr/bin/python

#       Sgourakis Lab
#   Author: Santrupti Nerli
#   Date: August 22, 2018
#   Email: snerli@ucsc.edu
#

'''

TCR class contains all the necessary functionalities required to fetch
necessary tcr sequences and their respective headers.

'''

from thread.fasta import FASTA

class TCR:

    # class members
    filename = "" # input file name containing tcr sequences in fasta format
    tcr_object = None # tcr object of FASTA type

    # constructor
    def __init__(self, filename):
        self.filename = filename
        self.tcr_object = FASTA(self.filename)
        self.tcr_object.read()

    # getter methods
    def get_headers(self):
        return self.tcr_object.get_headers()

    def get_sequences(self):
        return self.tcr_object.get_sequences()

    def get_sequence(self, header):
        return self.tcr_object.get_sequence(header)



#!/usr/bin/python

#       Sgourakis Lab
#   Author: Santrupti Nerli
#   Date: April 10, 2017
#   Email: snerli@ucsc.edu
#

# Load the Rosetta commands for use in the Python shell
from pyrosetta import *

# import rosetta files
from thread.template import TEMPLATE
from thread.thread import THREAD
from thread.fasta import FASTA
from thread.pre_thread import PRE_THREADING
from input_output.input.mhc import MHC
from input_output.input.peptides import PEPTIDE
from input_output.input.beta2m import BETA2M
from input_output.input.chaperone import CHAPERONE
from input_output.input.tcr import TCR

# import other required libraries
import os
import sys
import imp

'''

MODEL class contains all the necessary functionalities required to
invoke necessary functionailities to carry out the p/MHC-TCR/Chaperone threading.

'''

class MODEL:

    # class members
    complex_headers = [] # headers of the complex
    complex_sequences = [] # sequences of the complex
    template = None # template structure
    args = None # input arguments
    mpi_install = True # flag enabled if mpi is installed

    # constructor
    def __init__(self, args):
        self.args = args

    # method to model mhcs for given peptide, beta2m, tcr or chaperone sequences
    def model_mhc_for_each_peptide_beta2m_tcr_chaperone(self):
        mhc = MHC(self.args.get_mhcs(), self.args.is_no_trim_mhc_flag_set())
        beta2m = BETA2M(self.args.get_beta2m())
        peptides = PEPTIDE(self.args.get_peptides())
        chaperones = CHAPERONE(self.args.get_chaperone())
        tcr = TCR(self.args.get_tcr())

        if ( len(peptides.get_headers()) > 1 ):
            if ( self.args.get_pep_start_index() == 0 ):
                print ("Please provide peptide start index. Option: -pep_start_index")
                exit(1)
            if ( self.args.get_interface_cupoint() == 0 ):
                print ("Please provide interface cutpoint. Option: -interface_cutpoint")
                exit(1)
        '''
            # Not sure why I introduced this. Keeping it commented until I figure out that case

            elif ( len(peptides.get_headers()) == 1 ):
            for pep_header in peptides.get_headers():
                if ( pep_header is not "none" ):
                    print ("Please provide peptide start index. Option: -pep_start_index")
                    exit(1)
        '''

        # get all combinations of sequences
        # mhcs, beta2ms, peptides, tcrs and chaperones
        for mhc_header in mhc.get_headers():
            for beta2m_header in beta2m.get_headers():
                for pep_header in peptides.get_headers():
                    for tcr_header in tcr.get_headers():
                        for chaperone_header in chaperones.get_headers():
                            header = mhc_header+"_"+beta2m_header+"_"+pep_header+"_"+tcr_header+"_"+chaperone_header
                            sequence = mhc.get_sequence(mhc_header)+beta2m.get_sequence(beta2m_header)+peptides.get_sequence(pep_header)+tcr.get_sequence(tcr_header)+chaperones.get_sequence(chaperone_header)
                            self.generate_fasta(header, sequence)

        # identify the number of jobs
        njobs = len(self.complex_sequences)

        try:
            from jd.job_distributor import JOB_DISTRIBUTOR
            job_dist = JOB_DISTRIBUTOR()
            # Can we make this call by only once processor?
            '''
            template_name = job_dist.perform_single_operation(self.get_template)
            if self.template == None and isinstance(template_name, str):
                self.template = TEMPLATE(template_name).get_pose_from_pdb()
                job_dist.apply(njobs, self.apply)
            elif self.template != None:
                job_dist.apply(njobs, self.apply)
            '''
            self.get_template()
            #job_dist.SimpleMPIJobDistributor(njobs, self.apply)
            job_dist.MPIJobDistributor(njobs, self.apply)
        except ImportError:
            self.mpi_install = False

        # if not mpi then run this on a single processor
        if not self.mpi_install:
            self.get_template()
            for job_id in range(njobs):
                self.apply(job_id)
                job_id += 1

    # apply method that calls threader
    def apply(self, i):
        pre_thread = None
        pre_thread = PRE_THREADING(self.template, self.complex_sequences[i],
                            self.complex_headers[i], self.args)
        threader = THREAD(pre_thread)
        threader.apply()

    # method to initialize template structure
    def get_template(self):
        self.template = TEMPLATE(self.args.get_template_pdb())
        self.template.treat_template_structure(self.args.get_mhc_chain(),
                                            self.args.get_peptide_chain(),
                                            self.args.is_no_trim_mhc_flag_set(),
                                            self.args.get_mhc_trim_length(),
                                            self.args.get_idealize_relax())
        return self.template.get_pdb()

    # method to generate fasta file names
    def generate_fasta(self, header, sequence):
        fasta_file_name = header+".fasta"
        FASTA(fasta_file_name).write(header, sequence)
        self.complex_headers.append(header)
        self.complex_sequences.append(sequence)
        return fasta_file_name



#!/usr/bin/python

#       Sgourakis Lab
#   Author: Santrupti Nerli
#   Date: Jan 16, 2017
#   Email: snerli@ucsc.edu
#

'''

BLOSUM class contains all the necessary functionalities required to
obtain BLOSUM scoring matrix based on the sequence alignment identity.
For the purposes of MHC modeling we use BLOSUM62

'''

# bio libraries
from Bio.SubsMat.MatrixInfo import *

# import other required libraries
from collections import defaultdict

class BLOSUM:

    # class members
    BLOSUMX = defaultdict(dict) # variable to store the matrix

    # constructor
    # we support only three types of matrices
    # each with sequence identities 45%, 62% and 80%
    def __init__(self, type=62):
        if type == 62:
            blosum = blosum62
        elif type == 80:
            blosum = blosum80
        elif type == 45:
            blosum = blosum45
        else:
            print("Unknown substitution matrix")

        # fetch the matrix and convert it to dictionary format
        # for easier access
        for key in blosum:
            key0 = key[0]
            key1 = key[1]
            self.BLOSUMX[key0][key1] = blosum[key]
            self.BLOSUMX[key1][key0] = blosum[key]

        # provide scores to gaps denoted by asterisk
        starArr = {}
        for key in self.BLOSUMX:
            starArr[key] = -4
            starArr['*'] = 1

        for key in starArr:
            self.BLOSUMX['*'][key] = starArr[key]
            self.BLOSUMX[key]['*'] = starArr[key]

    # getter method
    def get_matrix(self):
        return self.BLOSUMX



#!/usr/bin/python

#       Sgourakis Lab
#   Author: Andrew C. McShan
#   Edited by: Santrupti Nerli
#   Date: June 20, 2018
#   Email: snerli@ucsc.edu
#

'''

A dictionary of supported beta2m sequences

'''
# import required libraries
from collections import defaultdict

beta2m_sequences = defaultdict(dict)
beta2m_sequences['humanbeta2m']='IQRTPKIQVYSRHPAENGKSNFLNCYVSGFHPSDIEVDLLKNGERIEKVEHSDLSFSKDWSFYLLYYTEFTPTEKDEYACRVNHVTLSQPKIVKWDRDM'
beta2m_sequences['mousebeta2m']='IQKTPQIQVYSRHPPENGKPNILNCYVTQFHPPHIEIQMLKNGKKIPKVEMSDMSFSKDWSFYILAHTEFTPTETDTYACRVKHASMAEPKTVYWDRDM'
beta2m_sequences['chickenbeta2m']='ADLTPKVQVYSRFPASAGTKNVLNCFAAGFHPPKISITLMKDGVPMEGAQYSDMSFNDDWTFQRLVHADFTPSSGSTYACKVEHETLKEPQVYKWDPEF'
beta2m_sequences['bovinebeta2m']='IQRPPKIQVYSRHPPEDGKPNYLNCYVYGFHPPQIEIDLLKNGEKIEKSEQSDLSFSKDWSFYLLSHAEFTPNSKDQYSCRVKHVTLEQPRIVKWDRDL'
beta2m_sequences['ratbeta2m']='IQKTPQIQVYSRHPPENGKPNFLNCYVSQFHPPQIEIELLKNGKKIPNIEMSDLSFSKDWSFYILAHTEFTPTETDVYACRVKHVTLKEPKTVTWDRDM'
beta2m_sequences['none']=''



from argparse import Namespace
import torch
import torch.nn as nn
import torch.nn.functional as F
import pytorch_lightning as pl
from torchmetrics import Accuracy
import math
from .utils import get_linear_schedule_with_warmup, aa_idx


class Attention(nn.Module):
    def __init__(self, atten_size, return_attention=False):
        super().__init__()
        self.w = nn.Parameter(torch.rand(atten_size, 1))
        self.b = nn.Parameter(torch.zeros(atten_size))
        self.return_attention = return_attention

    def forward(self, x: torch.Tensor):
        # [seq_len, batch, num_directions * hidden_size]
        assert len(x.size()) == 3
        assert self.w.size(0) == x.size(-1)
        elements = torch.matmul(x, self.w) + self.b
        elements = F.tanh(elements)  # [seq_len, batch, 1]
        alpha = F.softmax(elements, dim=0)  # [seq_len, batch, 1]
        out = x * alpha  # [seq_len, batch, num_directions * hidden_size]
        out = out.sum(dim=0)  # [batch, num_directions * hidden_size]
        assert out.size(0) == x.size(1)
        if self.return_attention:
            return out, alpha
        return out


class GRU_model(nn.Module):
    def __init__(self, config, dropout, bidirectional, use_attention=False):
        super().__init__()
        self.use_attention = use_attention
        self.n_direction = 2 if bidirectional else 1
        self.embed_dim = config['embed_dim']
        self.rnn_dim = config['rnn_dim']
        self.embedding = nn.Embedding(len(aa_idx), self.embed_dim)
        self.gru = nn.GRU(self.embedding.embedding_dim, self.rnn_dim, num_layers=3,
                          dropout=dropout, bidirectional=bidirectional)
        self.rnn_outsize = self.gru.hidden_size * self.n_direction
        self.linear = nn.Linear(self.rnn_outsize, 1)
        self.attention = Attention(atten_size=self.rnn_outsize)

    def forward(self, x):
        x = self.embedding(x)
        x = x.transpose(1, 0)
        output, hidden = self.gru(x)
        # output [seq_len, batch, num_directions * hidden_size]
        # hidden [num_layers * num_directions, batch, hidden_size]
        if self.n_direction == 2:
            # [batch, hidden_size * num_directions]
            hidden = torch.cat([hidden[-1], hidden[-2]], dim=1)
        else:
            hidden = hidden[-1]
        assert hidden.size(1) == self.n_direction*self.gru.hidden_size
        if self.use_attention:
            out = self.attention(output)
        else:
            out = hidden
        out = self.linear(out)
        out = torch.sigmoid(out)
        return out

    def get_rnn_feature(self, x, return_seq=False):
        x = self.embedding(x)
        x = x.transpose(1, 0)
        output, hidden = self.gru(x)
        # output [seq_len, batch, num_directions * hidden_size]
        # hidden [num_layers * num_directions, batch, hidden_size]
        if return_seq:
            return output.transpose(1, 0)
        if self.n_direction == 2:
            # [batch, hidden_size * num_directions]
            hidden = torch.cat([hidden[-1], hidden[-2]], dim=1)
        else:
            hidden = hidden[-1]
        assert hidden.size(1) == self.n_direction*self.gru.hidden_size
        if self.use_attention:
            out = self.attention(output)
        else:
            out = hidden
        return out


class PL_Sequence(pl.LightningModule):
    def __init__(self, config, total_steps, gpu=0):
        super().__init__()
        self.config = config
        self.total_steps = total_steps
        self.warmup_steps = math.ceil(total_steps * 0.1)
        self.gpu = gpu
        self.batch_size = config['batch_size']
        self.model_type = 'att_BGRU'
        self.dropout = config['dropout']
        self.train_acc = Accuracy()
        self.val_acc = Accuracy()
        self.save_hyperparameters(Namespace(**config))
        if self.model_type == 'GRU':
            self.model = GRU_model(
                config=config, dropout=self.dropout, bidirectional=False, use_attention=False)
        elif self.model_type == 'BGRU':
            self.model = GRU_model(
                config=config, dropout=self.dropout, bidirectional=True, use_attention=False)
        elif self.model_type == 'att_BGRU':
            self.model = GRU_model(
                config=config, dropout=self.dropout, bidirectional=True, use_attention=True)
        # print(self.model)

    def forward(self, x):
        out = self.model(x)
        return out

    def get_loss(self, pred, label):
        loss = nn.BCELoss()
        return loss(pred, label.float())

    def training_step(self, batch, batch_idx):
        input, label = batch
        pred = self.forward(input).squeeze(-1)
        loss = self.get_loss(pred, label)
        self.log(f'loss/train_loss', loss)
        self.train_acc(pred, label.int())
        self.log(f'acc/train_acc', self.train_acc,
                 on_step=False, on_epoch=True)
        return {'loss': loss}

    def validation_step(self, batch, batch_idx):
        input, label = batch
        pred = self.forward(input).squeeze(-1)
        loss = self.get_loss(pred, label)
        self.val_acc(pred, label.int())
        self.log('acc/val_acc', self.val_acc, on_step=False, on_epoch=True)
        return {'val_loss': loss}

    def validation_epoch_end(self, outputs):
        avg_loss = torch.stack([x["val_loss"] for x in outputs]).mean()
        self.log(f'loss/val_loss', avg_loss)

    def configure_optimizers(self):
        optimizer = torch.optim.AdamW(
            self.parameters(), lr=1e-3, weight_decay=1e-2)
        scheduler = get_linear_schedule_with_warmup(
            optimizer, num_warmup_steps=self.warmup_steps, num_training_steps=self.total_steps)
        return {'optimizer': optimizer, 'lr_scheduler': scheduler}



#!/usr/bin/python

#       Sgourakis Lab
#   Author: Santrupti Nerli
#   Date: August 22, 2018
#   Email: snerli@ucsc.edu
#

'''

HEADER_READER class contains all the necessary functionalities required to
read respective headers from the user input files.

'''

class HEADER_READER:

    # class members
    filename = "" # input file name containing beta2m, chaperone or mhc header lists
    headers = [] # header list

    # constructor
    def __init__(self, filename):
        self.filename = filename
        self.headers = []

    # method to read the file and populate headers
    # if filenames are not specified. For instance, chaperone,
    # then do nothing
    def read_headers(self):
        if self.filename != None:
            read_file_handler = open(self.filename, "r")
            for line in read_file_handler:
                line = line.rstrip()
                self.headers.append(line)
            read_file_handler.close()

    # getter method
    # if filename is not provided, then just return none
    def get_headers(self):
        if len(self.headers) == 0:
            self.headers.append('none')
        return self.headers



#!/usr/bin/python

#       Sgourakis Lab
#   Author: Santrupti Nerli
#   Date: July 9, 2018
#   Email: snerli@ucsc.edu
#

'''

RMSD class contains all the necessary functionalities required to compute RMSD values
of the threaded/relaxed models to the input native structure.

'''

# Load the Rosetta commands for use in the Python shell
from pyrosetta import *
from pyrosetta.rosetta.core.scoring import *

class RMSD:

    # class members
    pose = None # pose whose rmsd values are calculated
    native_pose = None # native pose
    residues_of_interest = [] # residues used to compute rmsds over

    # constructor
    def __init__(self, pose, native_pose, residues_of_interest = None):
        self.pose = pose
        self.native_pose = native_pose
        self.residues_of_interest = residues_of_interest

    def CA_rmsd(self):
        rms_ca = pyrosetta.rosetta.core.scoring.CA_rmsd(self.native_pose, self.pose)
        return rms_ca

    def CA_rmsd_between_indices(self, start, end):
        rms_ca = pyrosetta.rosetta.core.scoring.CA_rmsd(self.native_pose, self.pose, start, end)
        return rms_ca

    def CA_rmsd_residue_number(self, resi_vec):
        rms_ca = pyrosetta.rosetta.core.scoring.CA_rmsd(self.native_pose, self.pose, resi_vec)
        return rms_ca

    def all_atom_rmsd(self):
        rms_all = pyrosetta.rosetta.core.scoring.all_atom_rmsd(self.native_pose, self.pose)
        return rms_all



#!/usr/bin/python

#       Sgourakis Lab
#   Author: Santrupti Nerli
#   Date: May 25, 2017
#   Email: snerli@ucsc.edu
#

# Load the Rosetta commands for use in the Python shell
from pyrosetta import *

#custom libraries
from idealize_relax.relax import RELAX
from idealize_relax.movemap import MOVEMAP
from ia.chain_split import CHAIN_SPLIT
from ia.interface_analyzer import INTERFACE
from input_output.output.output_interface_energies import OUT_INTERFACE_ENERGY
from scoring.rmsd import RMSD

# import other required libraries
import os
import sys

'''

POST_THREADING class contains all the necessary functionalities required to
perform operations post threading process.

'''

class POST_THREADING:

    # class members
    args = None # store the input arguments
    threaded_pose = None # store the threaded pose
    tag = "" # tag is used for naming the output file
    pre_threader = None # pre threader object
    movemap = None # movemap object
    mpi_install = True # should I run mpi?
    output_energies = None # output the interface energies

    # constructor
    def __init__(self, pre_threader, threaded_pose, tag):
        self.args = pre_threader.args
        self.pre_threader = pre_threader
        self.threaded_pose = threaded_pose
        self.tag = tag
        self.output_energies = OUT_INTERFACE_ENERGY(self.args.get_out_file())

    # method to refine the structures post homology modeling
    def treatment_post_homology_modeling(self):
        # output threaded structure
        self.threaded_pose.dump_pdb(self.tag+".pdb")
        # create movemap
        self.movemap = MOVEMAP(self.tag+".pdb", self.args.get_pep_start_index(),
                            self.pre_threader.get_pep_length(), self.args.get_groove_distance(),
                            self.tag+".movemap")
        self.movemap.apply()

        # if refine after relax
        if self.args.get_relax_after_threading():

            relaxed_pose = self.threaded_pose

            # Can we split this into multiple nodes?
            # if we find more
            try:
                from jd.job_distributor import JOB_DISTRIBUTOR
                job_dist = JOB_DISTRIBUTOR()
                print ("No. of jobs for relax: ", self.args.get_nstruct())
                job_dist.SimpleMPIJobDistributor(self.args.get_nstruct(), self.minimize_and_calculate_energy)
            except ImportError:
                self.mpi_install = False

            # if mpi is not installed, just run on a
            # single thread
            if not self.mpi_install:
                for job_id in range(self.args.get_nstruct()):
                    self.minimize_and_calculate_energy(job_id)
            self.output_energies.write()

    # method to perform ebergy minimization and calculate
    # interface energies
    def minimize_and_calculate_energy(self, i):
        relax = RELAX()
        threaded_pose = Pose()
        threaded_pose.detached_copy(self.threaded_pose)

        # output relaxed structures
        relaxed_threaded_pose = Pose()
        relaxed_threaded_pose = relax.relax_pdb_with_movemap(threaded_pose, self.movemap.get_movemap())
        relaxed_tag = self.tag+"_relaxed_"+str(i)+".pdb"
        relaxed_threaded_pose.dump_pdb(relaxed_tag)

        # compute rmsd values to the native if provided
        all_rms = -1
        ca_rms = -1
        if self.args.get_native() != None:
            [all_rms, ca_rms] = self.report_rmsd(relaxed_tag)
            print("All atom RMSD: ", all_rms)
            print("CA RMSD: ", ca_rms)

        freshly_relaxed_pose = pose_from_pdb(relaxed_tag)
        # split the chains of the relaxed structures
        # according to the cutpoint because threader
        # create a singe chain
        chain_split = CHAIN_SPLIT(freshly_relaxed_pose, self.args.get_interface_cupoint())
        chain_split.cut()
        split_pose = Pose()
        split_pose = chain_split.get_pose()

        # calculate interface energies
        ia = INTERFACE(split_pose)
        ia.analyze()
        print("Interface energy: ", ia.get_dG())
        self.output_energies.add(self.tag, ia.get_dG(), all_rms, ca_rms)

    def report_rmsd(self, tag, movemap = None):
        pose = pose_from_pdb(tag)
        native_pose = pose_from_pdb(self.args.get_native())
        rms_obj = RMSD(pose, native_pose)

        all = rms_obj.all_atom_rmsd()
        if movemap != None:
            ca = rms_obj.CA_rmsd(movemap)
        else:
            ca = rms_obj.CA_rmsd()

        return [all, ca]



#!/usr/bin/python

#       Sgourakis Lab
#   Author: Santrupti Nerli
#   Date: December 1, 2017
#   Email: snerli@ucsc.edu
#

# Load the Rosetta commands for use in the Python shell
from pyrosetta import *

# import rosetta files
from pyrosetta.rosetta.protocols.comparative_modeling import *
from pyrosetta.rosetta.core.scoring import *
from pyrosetta.rosetta.core.sequence import *

#custom libraries
from thread.post_thread import POST_THREADING

# import other required libraries
import os
import sys

'''

THREAD class contains all the necessary functionalities required to
perform target onto the template structure.

'''

class THREAD:

    # class members
    pre_threader = None # pre threader object

    # constructor
    def __init__(self, pre_thread):
        self.pre_threader = pre_thread

    # method that calls threader
    def apply(self):
        if self.pre_threader != None:
            self.do_partial_threading()

    # method to perform partial threading
    def do_partial_threading(self):
        # for each alignment combination in the grishin file.
        alignment_vec = read_aln("grishin", self.pre_threader.get_grishin_file_name())
        for vec in alignment_vec:
            new_pose = pose_from_sequence(self.pre_threader.get_target_sequence(), "fa_standard")
            thread = PartialThreadingMover(vec, self.pre_threader.get_template().get_pose())

            thread.apply(new_pose)

            threaded_pose = new_pose
            tag = self.pre_threader.get_target_file_name()

            post = POST_THREADING(self.pre_threader, threaded_pose, tag)
            post.treatment_post_homology_modeling()



import os
import sys
import numpy
import argparse
from collections import defaultdict

'''
python ~/rosettamhc/mhc-pep-threader/covid/identify_common_cold_peps.py -e sars_eptiopes.txt

'''

# get command line arguments
def get_args():
    """
    Parse command line arguments
    """

    parser = argparse.ArgumentParser(description="Method to setup rosettamhc jobs")
    parser.add_argument("-e", help="sars eptiopes, one line for each epitope in the text file")
    parser.add_argument("-aln", help="alignment file")
    parser.add_argument("-tag", help="protein name")
    parser.add_argument("-length", help="epitope length", default=9, type=int)
    parser.add_argument("-name", help="protein name [229E, NL63, OC43, HKU1]")

    args = parser.parse_args()
    return args

# read SARS-CoV-2 epitope file
def read_epitopes(args):

    inputfilehandler = open(args.e, 'r')
    epitopes = []
    for line in inputfilehandler:
        line = line.rstrip()
        epitopes.append(line)
    inputfilehandler.close()

    return epitopes

# perform sliding window
def sliding_window(seq, args):

    sequence = list(seq)
    epitopes = {}
    repeat = 0
    length = args.length

    for i in range(0, len(sequence)-length+1):
        pep = ''.join(sequence[i:i+length])
        key = pep
        if key not in epitopes:
            epitopes[pep] = i
        else:
            repeat += 1

        #print ("Found repeats: ", repeat)

    return epitopes

# read alignment file
def read_aln(args):

    inputfilehandler = open(args.aln, 'r')
    sarscov2 = ''
    hcov229e = ''
    hcovnl63 = ''
    hcovoc43 = ''
    hcovhku1 = ''
    for line in inputfilehandler:
        line = line.rstrip()
        if "CLUSTAL" not in line and line != '' and '*' not in line and ':' not in line:
            fields = line.split()
            if '229E_strain' in line:
                hcov229e += fields[1]
            elif 'NL63_strain' in line:
                hcovnl63 += fields[1]
            elif 'SARS_COV2' in line:
                sarscov2 += fields[1]
            elif 'OC43_strain' in line:
                hcovoc43 += fields[1]
            elif 'HKU1_strain' in line:
                hcovhku1 += fields[1]
    inputfilehandler.close()

    return (sarscov2, hcov229e, hcovnl63, hcovoc43, hcovhku1)

def print_epitopes(e, match, args, strain):

    if '-' not in match:
        print ('>'+e+'_'+match+'_'+args.tag+'_'+strain)
        print (match)

def match(epitopes, new_epitopes, hcov229e, hcovnl63, hcovoc43, hcovhku1, args):

    hcov229e_arr = list(hcov229e)
    hcovnl63_arr = list(hcovnl63)
    hcovoc43_arr = list(hcovoc43)
    hcovhku1_arr = list(hcovhku1)

    specific = {}

    print ("SARS-CoV-2 cross-reactive epitopes:")
    for e in epitopes:
        e_len = len(list(e))
        if e in new_epitopes:
            hcov229e_match = ''.join(hcov229e_arr[new_epitopes[e]:new_epitopes[e]+e_len])
            hcovnl63_match = ''.join(hcovnl63_arr[new_epitopes[e]:new_epitopes[e]+e_len])
            hcovoc43_match = ''.join(hcovoc43_arr[new_epitopes[e]:new_epitopes[e]+e_len])
            hcovhku1_match = ''.join(hcovhku1_arr[new_epitopes[e]:new_epitopes[e]+e_len])

            if args.name == '229E':
                print_epitopes(e, hcov229e_match, args, "229E")
            elif args.name == 'NL63':
                print_epitopes(e, hcovnl63_match, args, "NL63")
            elif args.name == 'OC43':
                print_epitopes(e, hcovoc43_match, args, "OC43")
            elif args.name == 'HKU1':
                print_epitopes(e, hcovhku1_match, args, "HKU1")

            if '-' in hcov229e_match or '-' in hcovhku1_match or '-' in hcovnl63_match or '-' in hcovoc43_match:
                specific[e] = e+','+args.tag+','+hcov229e_match+','+hcovhku1_match+','+hcovnl63_match+','+hcovoc43_match
            #print ("Found "+e, hcov229e_match, hcovnl63_match, hcovoc43_match, hcovhku1_match)

    print ("SARS-CoV-2 specific epitopes:")
    for pep in specific:
        print (specific[pep])

# main method
def main(args):

    epitopes = read_epitopes(args)
    (sarscov2, hcov229e, hcovnl63, hcovoc43, hcovhku1) = read_aln(args)
    new_epitopes = sliding_window(sarscov2, args)
    match(epitopes, new_epitopes, hcov229e, hcovnl63, hcovoc43, hcovhku1, args)

# Entry to the program
if __name__ == "__main__":

    main(get_args())



#!/usr/bin/python

#       Sgourakis Lab
#   Author: Santrupti Nerli
#   Date: January 26, 2017
#   Email: snerli@ucsc.edu
#

'''

RELAX class contains all the necessary functionalities required to call
the appropriate minimizer for a given pose.

'''

# Load the Rosetta commands for use in the Python shell
from pyrosetta import *

# import rosetta files
from pyrosetta.rosetta.protocols.relax import *
from pyrosetta.rosetta.core.scoring import *

class RELAX:

    # method to perform simple minimization using FastRelax
    def relax_pdb(self, pose):
        relax = FastRelax()
        relax.set_scorefxn(get_fa_scorefxn())
        relax.apply(pose)
        return pose

    # method to perform simple minimization using FastRelax
    # also minimizes only those residues specified in the movemap
    def relax_pdb_with_movemap(self, pose, movemap):
        relax = FastRelax()
        relax.set_scorefxn(get_fa_scorefxn())
        relax.set_movemap(movemap)
        relax.apply(pose)
        return pose



#!/usr/bin/python

#       Sgourakis Lab
#   Author: Santrupti Nerli
#   Date: February 2, 2018
#   Email: snerli@ucsc.edu
#

'''

OUT_INTERFACE_ENERGY class contains all the necessary functionalities required
create output file containing targets and their binding energies.

'''

# import required libraries
from collections import defaultdict

class OUT_INTERFACE_ENERGY:

    # class members
    filename = "" # filename to write the output to
    energies = None # dictionary to store the binding energies

    # constructor
    def __init__(self, filename):
        self.energies = defaultdict(list)
        self.filename =  filename

    # method to populate dictionary with binding energies
    def add(self, key, value, rms1 = -1, rms2 = -1):
        fields = key.split("_")[:4]
        if rms1 == -1 and rms2 == -1:
            self.energies[key].append(value)
        else:
            metrics = []
            metrics.append(value)
            metrics.append(rms1)
            metrics.append(rms2)
            self.energies[key].append(metrics)

    # method to output a csv file containing names of domains
    # and their corresponding binding energies
    def write(self):
        write_file_handler = open(self.filename, "a")
        for key,value in self.energies.items():
            fields = key.split("_")
            mhc = fields[0]
            beta2m = fields[1]
            pep = fields[2]
            tcr = fields[3]
            chaperone = fields[4]
            energy = ""
            for i in range(len(value)):
                energy += str(value[i])
                if i != len(value)-1:
                    energy += ","
            text = mhc+","+beta2m+","+pep+","+tcr+","+chaperone+","+energy+"\n"
            write_file_handler.write(text)
            print (text)
        write_file_handler.close()



#!/usr/bin/python

#       Sgourakis Lab
#   Author: Santrupti Nerli
#   Date: June 20, 2018
#   Email: snerli@ucsc.edu
#

'''

A dictionary of supported chaperone sequences

'''
# import required libraries
from collections import defaultdict

chaperone_sequences = defaultdict(dict)
chaperone_sequences['tapasin']='MKSLSLLLAVALGLATAVSAGPAVIECWFVEDASGKGLAKRPGALLLRQGPGEPPPRPDLDPELYLSVHDPAGALQAAFRRYPRGAPAPHCEMSRFVPLPASAKWASGLTPAQNCPRALDGAWLMVSISSPVLSLSSLLRPQPEPQQEPVLITMATVVLTVLTHTPAPRVRLGQDALLDLSFAYMPPTSEAASSLAPGPPPFGLEWRRQHLGKGHLLLAATPGLNGQMPAAQEGAVAFAAWDDDEPWGPWTGNGTFWLPRVQPFQEGTYLATIHLPYLQGQVTLELAVYKPPKVSLMPATLARAAPGEAPPELLCLVSHFYPSGGLEVEWELRGGPGGRSQKAEGQRWLSALRHHSDGSVSLSGHLQPPPVTTEQHGARYACRIHHPSLPASGRSAEVTLEVAGLSGPSLEDSVGLFLSAFLLLGLFKALGWAAVYLSTCKDSKKKAE'
chaperone_sequences['tapbpr']='MGTQEGWCLLLCLALSGAAETKPHPAEGQWRAVDVVLDCFLAKDGAHRGALASSEDRARASLVLKQVPVLDDGSLEDFTDFQGGTLAQDDPPIIFEASVDLVQIPQAEALLHADCSGKEVTCEISRYFLQMTETTVKTAAWFMANMQVSGGGPSISLVMKTPRVTKNEALWHPTLNLPLSPQGTVRTAVEFQVMTQTQSLSFLLGSSASLDCGFSMAPGLDLISVEWRLQHKGRGQLVYSWTAGQGQAVRKGATLEPAQLGMARDASLTLPGLTIQDEGTYICQITTSLYRAQQIIQLNIQASPKVRLSLANEALLPTLICDIAGYYPLDVVVTWTREELGGSPAQVSGASFSSLRQSVAGTYSISSSLTAEPGSAGATYTCQVTHISLEEPLGASTQVVPPERRTALGVIFASSLFLLALMFLGLQRRQAPTGLGLLQAERWETTSCADTQSSHLHEDRTARVSQPS'
chaperone_sequences['none']=''



#!/usr/bin/python

#       Sgourakis Lab
#   Author: Santrupti Nerli
#   Date: December 1, 2017
#   Email: snerli@ucsc.edu
#

'''

MOVEMAP class contains all the necessary functionalities required to create a
a set of residues whose backbone and side-chain torsion angles are allowed to
modify during structure refinement.

'''

# import other required libraries
import sys
import math

# import biopython libraries
#
# Hamelryck, T., Manderick, B. (2003) PDB parser and structure class implemented in Python. Bioinformatics 19: 2308-2310
from Bio.PDB import *
#

# Rosetta libraries
from pyrosetta.rosetta.core.kinematics import MoveMap

# custom libraries
from util.custom_pdb import MyPDB


class MOVEMAP:

    # class members
    structure = None # store the pdb file
    pep_start_index = None # store the starting residue index from which you want to create movemap
    pep_len = -1 # length of a peptide
    groove_distance = 3.5 # also include those residues that are at a distance of 3.5 A from peptide
    filename = "" # movemap file name

    protein_residues = []
    peptide_residues = []
    movemap_res_list = []

    # constructor
    def __init__(self, pdb_file, pep_start_index, pep_length, groove_distance, filename):
        self.structure = MyPDB(pdb_file)
        self.structure.read_chain_A() # This should work since movemap is created after threading is done which creates only a single chain A
        self.pep_start_index = pep_start_index
        self.groove_distance = groove_distance
        self.filename = filename
        self.pep_len = pep_length

    # method to create a write to movemap file
    def apply(self):
        self.protein_residues = []
        self.peptide_residues = []
        self.movemap_res_list = []
        self.extract_peptide_protein_residues()
        self.create_pep_protein_list()
        self.write_movemap()

    # method to make a list of all peptide and protein residues that
    # will be filtered in the next stage
    def extract_peptide_protein_residues(self):
        residue_number = 1
        print (self.pep_start_index+self.pep_len)
        for residue in self.structure.get_pdb():
            if residue_number >= self.pep_start_index and residue_number < self.pep_start_index+self.pep_len:
                self.peptide_residues.append(residue)
            else:
                self.protein_residues.append(residue)
            residue_number += 1

    # method to fetch all the residues that can be moved during refinement
    def create_pep_protein_list(self):
        for pep_res in self.peptide_residues: # for each peptide residue
            for protein_res in self.protein_residues: # for each protein residue
                for pep_atom in pep_res: # for each atom in the pep residue
                    for protein_atom in protein_res: # for each atom in the protein residue
                        distance = math.fabs(pep_res[pep_atom.get_name()] - protein_res[protein_atom.get_name()]) # calculate distance between two atoms
                        if distance <= self.groove_distance:
                            residue_number = int(str(protein_res).split()[3].split("=")[1])
                            if residue_number not in self.movemap_res_list:
                                self.movemap_res_list.append(residue_number)
            self.movemap_res_list.append(int(str(pep_res).split()[3].split("=")[1]))

    # method to write to a movemap file
    # see here for movemap file requirements: https://www.rosettacommons.org/docs/latest/rosetta_basics/file_types/movemap-file
    def write_movemap(self):
        movemap_handler = open(self.filename, "w")
        movemap_handler.write("RESIDUE * NO\n")
        for residue in self.movemap_res_list:
            movemap_handler.write("RESIDUE "+str(residue)+" BBCHI\n")
        movemap_handler.write("JUMP * YES\n")
        movemap_handler.close()

    # getter methods
    def get_filename(self):
        return self.filename

    # method to create Rosetta specific movemap object
    # read from file and return thr object
    def get_movemap(self):
        movemap_object = MoveMap()
        movemap_object.init_from_file(self.filename)
        return movemap_object

    def extract_residues_from_movemap(movemap):
        readfilehandler = open(movemap, "r")
        resi_list = pyrosetta.rosetta.std.list_unsigned_long_t()
        for line in readfilehandler:
            line = line.rstrip()
            if "RESIDUE" in line and "CHI" in line:
                resi = line.split(" ")[1]
                resi_list.append(int(resi))
        readfilehandler.close()
        return resi_list



#!/usr/bin/python

#       Sgourakis Lab
#   Author: Santrupti Nerli
#   Date: January 26, 2017
#   Email: snerli@ucsc.edu
#

'''

ALIGN class contains all the necessary functionalities required to perform sequence
alignment between the template and the target sequences.

'''

# additional bio libraries
# Refernce: Fast, scalable generation of high-quality protein multiple sequence alignments using Clustal Omega.
# (2011 October 11) Molecular systems biology 7 :539
from Bio.Align.Applications import ClustalOmegaCommandline

#custom libraries
from alignment.substitution_matrix import BLOSUM

# import other required libraries
import os
import sys
import subprocess

class ALIGN:

    # class members
    template_seq = None # amino acid sequence for a given template
    target_header = None
    target_seq = None # amino acid sequence for the required target
    matrix_type = 62 # matrix type to use for scoring alignment
    clustal_input = "" # input filename for clustal omega
    clustal_output = "" # output filename for clustal omega
    clustal_path = 'clustalo'

    # constructor
    def __init__(self, template_seq, target_header, target_seq, clustal_path, matrix_type = 62):
        self.template_seq = template_seq
        self.matrix_type = matrix_type
        self.target_seq = target_seq
        self.clustal_input = target_header+"_clustal_input.fasta"
        self.clustal_output = target_header+"_clustal_output.fasta"
        self.clustal_path = clustal_path

    # method to create clustal input file
    def init_clustal_input(self):
        writefilehandle = open(self.clustal_input, "w")
        writefilehandle.write(">template\n")
        writefilehandle.write(self.template_seq+"\n")
        writefilehandle.write(">target\n")
        writefilehandle.write(self.target_seq+"\n")
        writefilehandle.close()

    # method that calls externally installed clustal program
    # read the input sequences from clustal input file and write to the
    # clustal output file
    def clustal(self):
        self.init_clustal_input()
        cline = ClustalOmegaCommandline(cmd=self.clustal_path, infile=self.clustal_input, outfile=self.clustal_output, distmat_full=True,
                                        verbose=True, seqtype="Protein", outfmt="vienna", iterations=10, percentid=True, force=True)
        cline()

    # getter method
    def get_clustal_output_filename(self):
        return self.clustal_output



'''

Method to read multiple fasta files in a directory and create epitopes of given lengths

'''


import os
import sys
import argparse
from thread.fasta import FASTA

# read arguments
def get_args():
    """
    Parse command line arguments
    """

    parser = argparse.ArgumentParser(description="Method to read multiple fasta files and create epitope list of given lengths")
    parser.add_argument("-i", "--infile", help="fully qualified path to the fasta file")
    parser.add_argument("-l", "--length", type=int, default=9, help="Epitope length")

    args = parser.parse_args()

    return args

# perform sliding window and extract the epitopes of specified lengths
def sliding_window(args, fasta):

    headers = fasta.get_headers()
    epitopes = {}

    repeat = 0
    for header in headers:
        sequence = list(fasta.get_sequence(header))
        for i in range(0, len(sequence)-args.length+1):
            pep = ''.join(sequence[i:i+args.length])
            key = '>'+pep
            if key not in epitopes:
                epitopes[key] = pep
            else:
                repeat += 1

    for pep in epitopes:
        print (pep+"\n"+epitopes[pep])

    #print ("Total:", len(epitopes), "Repeats: ", repeat)

# main method
def main():

    args = get_args()

    fasta = FASTA(args.infile)
    fasta.read()

    sliding_window(args, fasta)


# Entry to the program
if __name__ == "__main__":
    main()



#!/usr/bin/python

#       Sgourakis Lab
#   Author: Santrupti Nerli
#   Date: January 26, 2017
#   Email: snerli@ucsc.edu
#

'''

IDEALIZE class contains all the necessary functionalities required to idealize
an input pose.

'''

# Load the Rosetta commands for use in the Python shell
from pyrosetta import *

# import rosetta files
from pyrosetta.rosetta.protocols.idealize import *

class IDEALIZE:

    # method to apply IdealizeMover for a given pose
    def idealize_pdb(self, pose):
        ideal = IdealizeMover()
        ideal.apply(pose)
        return pose



#!/usr/bin/python

#       Sgourakis Lab
#   Author: Santrupti Nerli
#   Date: February 1, 2018
#   Email: snerli@ucsc.edu
#

'''

CHAIN_SPLIT class contains all the necessary functionalities required to divide
a pose into two separate chains based on the given residue number also referred to as cutpoint.

'''

# Load the Rosetta commands for use in the Python shell
from pyrosetta import *
from pyrosetta.rosetta.protocols.simple_moves import ChainSplitMover

class CHAIN_SPLIT:

    # class members
    pose = None # pose which needs to be split into two chains based on the residue number
    cutpoint = 1 # residue number that marks the split between two chains

    # constructor
    def __init__(self, pose, cutpoint):
        self.pose = pose
        self.cutpoint = cutpoint

    # method to perform chain split
    def cut(self):
        split = ChainSplitMover(self.cutpoint)
        split.apply(self.pose)

    # getter method
    def get_pose(self):
        return self.pose



"""
* Software License Agreement (BSD License)
*
*  Copyright (c) 2019, Rice University.
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without
*  modification, are permitted provided that the following conditions
*  are met:
*
*   * Redistributions of source code must retain the above copyright
*     notice, this list of conditions and the following disclaimer.
*   * Redistributions in binary form must reproduce the above
*     copyright notice, this list of conditions and the following
*     disclaimer in the documentation and/or other materials provided
*     with the distribution.
*   * Neither the name of the Rice University nor the names of its
*     contributors may be used to endorse or promote products derived
*     from this software without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
*  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
*  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
*  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
*  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
*  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
*  POSSIBILITY OF SUCH DAMAGE.
*********************************************************************/

Author: Jayvee Abella
"""

import sys
import os
from subprocess import call
from subprocess import check_output
from threading import Thread
import numpy as np
import mdtraj as md
import argparse
import glob
import time
from simtk.openmm.app import *
from simtk.openmm import *
from simtk.unit import *
from sys import stdout

one_letter_code = {'ARG':'R', 'HIS':'H', 'LYS':'K', 'ASP':'D', 'GLU':'E', \
                  'SER':'S', 'THR':'T', 'ASN':'N', 'GLN':'Q', 'CYS':'C', \
                  'GLY':'G', 'PRO':'P', 'ALA':'A', 'VAL':'V', 'ILE':'I', \
                  'LEU':'L', 'MET':'M', 'PHE':'F', 'TYR':'Y', 'TRP':'W'}

three_letter_code  = {v: k for k, v in one_letter_code.items()}

defaults_location = os.path.dirname(os.path.abspath(__file__)) #sys.path[0]
pymol_location = "pymol"
RCD_location = "rcd"
smina_location = "smina"
vina_location = "vina_split"
pdbfixer_location = "pdbfixer"

def main(args):

    parser = argparse.ArgumentParser(description="Anchored Peptide-MHC Ensemble Generator", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('peptide_input', type=str, nargs=1, help='Sequence of peptide to dock or pdbfile of crystal structure')
    parser.add_argument('receptor_class', type=str, nargs=1, help='Class descriptor of MHC receptor. Use REDOCK along with crystal input to perform redocking. Or pass a PDB file with receptor')
    
    parser.add_argument('--output_dir', type=str)
    parser.add_argument('--filename', type=str)
    parser.add_argument('--scoring', type=str, default='vinardo')
    parser.add_argument("--save_interval", type=int, default=5, help='Openmm simulation save interval (picosecond)')
    parser.add_argument("--step_size", type=float, default=0.002, help='Step size for Openmm Integrator if needed (picosecond)')
    parser.add_argument("--total_time", type=int, default=5000, help='Openmm simulation time (picosecond)')

    parser.add_argument("-n", "--num_cores", type=int, default=8, help='Number of cores to use for RCD and smina computations.')
    parser.add_argument("-l", "--num_loops", type=int, default=100, help='Number of loops to generate with RCD. (Note that the final number of sampled conformations may be less due to steric clashes.')
    parser.add_argument("-t", "--RCD_dist_tol", type=float, default=1.0, help='RCD tolerance (in angstroms) of inner residues when performing IK')
    parser.add_argument("-r", "--rigid_receptor", action="store_true", help='Disable sampling of receptor degrees of freedom specified in flex_res.txt')
    parser.add_argument("-d", "--debug", action="store_true", help='Print extra information for debugging')
    parser.add_argument("-p", "--save_only_pep_confs", action="store_true", help='Disable saving full conformations (peptide and MHC)')
    #parser.add_argument("--n-mer-templates", default="", help='File with n-mer pdb templates.')
    #parser.add_argument("--receptor-class-templates", default="", help='File with pdb receptor class templates')
    #parser.add_argument("--flex_res", default="", help='File with flexible residues')
    parser.add_argument("-a", "--anchor_tol", type=float, default=2.0, help='Anchor tolerance (in angstroms) of first and last backbone atoms of peptide when filtering')
    parser.add_argument("-o", "--score_with_openmm", action="store_true", help='Rescore full conformations with openmm (AMBER)')
    parser.add_argument("-g", "--num_rounds", type=int, default=1, help='Number of rounds to perform.')
    parser.add_argument("-b", "--pass_type", type=str, default='receptor_only', choices=['receptor_only', 'pep_and_recept'], help="When using multiple rounds, pass best scoring conformation across different rounds (choose either 'receptor_only' or 'pep_and_recept')")
    parser.add_argument("-s", "--min_with_smina", action="store_true", help='Minimize with SMINA instead of the default Vinardo')
    parser.add_argument("--use_gpu", action="store_true", help='Use GPU for OpenMM Minimization step')
    parser.add_argument("--no_progress", action="store_true", help='Do not print progress bar')
    parser.add_argument("--clean_rcd", action="store_true", help='Remove RCD folder at the end of each round')

    args = parser.parse_args(args)

    peptide_input = args.peptide_input[0]
    receptor_class = args.receptor_class[0]
    
    filename = args.filename
    output_dir = os.path.realpath(args.output_dir)
    if not os.path.exists(output_dir):
        os.mkdir(output_dir)
    scoring = args.scoring
    
    save_interval = args.save_interval
    step_size = args.step_size
    total_time = args.total_time

    num_cores = args.num_cores
    num_loops = args.num_loops
    RCD_dist_tol = args.RCD_dist_tol
    doReceptorMinimization = not args.rigid_receptor
    debug = args.debug
    saveFullConfs = not args.save_only_pep_confs
    anchor_tol = args.anchor_tol
    #anchor_tol = 3.0
    score_with_openmm = args.score_with_openmm
    num_rounds = args.num_rounds
    pass_type = args.pass_type
    min_with_smina = args.min_with_smina
    use_gpu = args.use_gpu
    if use_gpu: device = "OpenCL"
    else: device = "CPU"
    no_progress = args.no_progress
    printProgress = not no_progress
    cleanRCD = args.clean_rcd

    # print("Preparing peptide and MHC")

    native_loc = None
    performRedock = False
    # if input is PDB, then use it to define 'native' for later comparison
    if peptide_input[-4:] == ".pdb":

        native_pdb = peptide_input

        call(["grep \"[A-Z] C  \" " + native_pdb + " > native.pdb"], shell=True)
        native_loc = "native.pdb"
        f = md.load("native.pdb")
        seq_arr = [one_letter_code[r.name] for r in f.top.residues]
        seq = ''
        for s in seq_arr: seq += s
        peptide_sequence = seq
        print("Peptide sequence: " + peptide_sequence)
        print("Receptor class: " + receptor_class)

        if receptor_class == "REDOCK":
            performRedock = True
            receptor_template = native_pdb

    # otherwise, only sequence is needed
    else:
        peptide_sequence = peptide_input
        for resi in peptide_sequence:
            if resi not in list(three_letter_code.keys()):
                print("Unrecognized one letter residue code: " + resi)
                sys.exit(0)
        # print("Peptide sequence: " + peptide_sequence)
        # print("Receptor class: " + receptor_class)

        if receptor_class == "REDOCK":
            print("Must input native pdb as peptide input in order to perform redocking.")
            sys.exit(0)

    # retrieve peptide template
    n_mer_templates = {}
    f = open(defaults_location + "/n-mer-templates.txt")
    for line in f:
        line_arr = line.split()
        n_mer_templates[line_arr[0]] = line_arr[1]
    f.close()
    if str(len(peptide_sequence)) in list(n_mer_templates.keys()):
        peptide_template = n_mer_templates[str(len(peptide_sequence))]
    else:
        print(str(len(peptide_sequence)) + "-mers are not implemented yet. Check n-mer-templates.txt")
        sys.exit(0)

    call(["cp " + defaults_location + "/templates/" + peptide_template + " ."], shell=True) 

    # if pdb is provided, use MHC inside
    # elseif receptor allotype is specified, read in receptor templates
    if receptor_class[-4:] == ".pdb": receptor_template = receptor_class
    elif not performRedock:
        receptor_class_templates = {}
        f = open(defaults_location + "/receptor-class-templates.txt")
        for line in f:
            line_arr = line.split()
            receptor_class_templates[line_arr[0]] = line_arr[1]
        f.close()
        if receptor_class in list(receptor_class_templates.keys()):
            receptor_template = receptor_class_templates[receptor_class]
        else:
            print(receptor_class + " not yet implemented as a receptor class. Check receptor-class-templates.txt")
            sys.exit(0)
        call(["cp " + defaults_location + "/templates/" + receptor_template + " ."], shell=True)

    # read in flexible residues
    flexible_residues = ""
    f = open(defaults_location + "/flex_res.txt")
    for line in f:
        flexible_residues += line
    f.close()
    flexible_residues = flexible_residues.rstrip()

    for current_round in range(num_rounds): 

        if not os.path.exists(str(current_round)): call(["mkdir " + str(current_round)], shell=True)
        os.chdir(str(current_round))

        if current_round == 0:
            call(["cp ../*.pdb ."], shell=True)
        else:
            call(["cp ../*.pdb ."], shell=True)
            if pass_type == 'receptor_only':
                call(["cp ../" + str(current_round-1) + "/min_energy_system.pdb ."], shell=True)
                receptor_template = "min_energy_system.pdb"
            elif pass_type == 'pep_and_recept':
                call(["cp ../" + str(current_round-1) + "/min_energy_system.pdb ."], shell=True)
                receptor_template = "min_energy_system.pdb"
                peptide_template = "min_energy_system.pdb"
            call(["sed -i \"s/          /           /g\" min_energy_system.pdb"], shell=True) # fix annoying openbabel warning

        # print("Aligning peptide anchors to MHC pockets")

        call([pymol_location + " -qc " + defaults_location + "/align.py " + peptide_template + " " + receptor_template + " >> align.log 2>&1"], shell=True)

        if native_loc != None:
            call([pymol_location + " -qc " + defaults_location + "/align.py " + native_pdb + " " + receptor_template + " >> align.log 2>&1"], shell=True)
            call(["grep \"[A-Z] C  \" aln-" + native_pdb + " > native.pdb"], shell=True)
            native_loc = "native.pdb"

        call(["cp " + receptor_template + " ./receptor.pdb; sed -i \"/[A-Z] C/d\" receptor.pdb"], shell=True)

        call(["grep \"[A-Z] C  \" aln-" + peptide_template + " > peptide_template.pdb"], shell=True)
        f = md.load("peptide_template.pdb")
        seq_arr = [r.name for r in f.top.residues]
        call(["sed -i \"s/" + seq_arr[0] + " C   1/" + three_letter_code[peptide_sequence[0]] + " C   1/g\" peptide_template.pdb"], shell=True)
        call(["sed -i \"s/" + seq_arr[1] + " C   2/" + three_letter_code[peptide_sequence[1]] + " C   2/g\" peptide_template.pdb"], shell=True)

        if len(peptide_sequence) < 10:
            last_anchor_resi_1 = " C   " + str(len(peptide_sequence) - 1)
            last_anchor_resi_2 = " C   " + str(len(peptide_sequence))
        elif len(peptide_sequence) == 10:
            last_anchor_resi_1 = " C   " + str(len(peptide_sequence) - 1)
            last_anchor_resi_2 = " C  " + str(len(peptide_sequence))
        elif len(peptide_sequence) > 10:
            last_anchor_resi_1 = " C  " + str(len(peptide_sequence) - 1)
            last_anchor_resi_2 = " C  " + str(len(peptide_sequence))

        call(["sed -i \"s/" + seq_arr[-2] + last_anchor_resi_1 + "/" + three_letter_code[peptide_sequence[-2]] + last_anchor_resi_1 + "/g\" peptide_template.pdb"], shell=True)
        call(["sed -i \"s/" + seq_arr[-1] + last_anchor_resi_2 + "/" + three_letter_code[peptide_sequence[-1]] + last_anchor_resi_2 + "/g\" peptide_template.pdb"], shell=True)
        if len(peptide_sequence) < 12:
            call(["sed -i \"/C   [3-" + str(len(peptide_sequence) - 2) + "]/d\" peptide_template.pdb"], shell=True)
        else:
            call(["sed -i \"/C   [3-9]/d\" peptide_template.pdb"], shell=True)
            for j in range(len(peptide_sequence)-1 - 10):
                call(["sed -i \"/C  1" + str(j) + "/d\" peptide_template.pdb"], shell=True)

        f = open("peptide_template.pdb", 'r')
        f_new = open("anchors.pdb", "w")
        for line in f:
            firstword = line.split()[0]
            elem = line.split()[2]
            resi = line.split()[3]
            if firstword == 'ATOM':
                if resi != 'GLY':
                    if elem in ['N', 'CA', 'C', 'O', 'CB']:
                        f_new.write(line)
                else:
                    if elem in ['N', 'CA', 'C', 'O']:
                        f_new.write(line)

        f.close()
        f_new.close()
        call(["cp anchors.pdb peptide.pdb"], shell=True)

        # print("Sampling peptide backbone")
        if os.path.exists("RCD"):
            print("Found RCD folder. Skipping this step")
            ref = md.load("confs_top.pdb")
            ref_top = ref.top
            if native_loc != None:
                print("Native crystal found!")
                native = get_conf(native_loc, ref_top, "element != H", debug)
                print(native)
            else: native = None
            call(["cp confs_top.pdb peptide.pdb"], shell=True)
            call(["sed -i \"s/          /           /g\" peptide.pdb"], shell=True) # fix annoying openbabel warning

        else:
            call(["mkdir -p RCD/input"], shell=True)
            os.chdir("RCD/input")
            
            
            call(["cp ../../receptor.pdb anchored_pMHC.pdb"], shell=True)
            call(["cat ../../anchors.pdb >> anchored_pMHC.pdb"], shell=True)

            call(["cp " + defaults_location + "/loco.score " + defaults_location + "/dunbrack.bin ."], shell=True)

            last_loop_residue_1index = len(peptide_sequence) - 2
            call(["echo \"anchored_pMHC.pdb 3 " + str(last_loop_residue_1index) + " C " + peptide_sequence[2:last_loop_residue_1index] + "\" > loops.txt"], shell=True)

            #call([mpi_location + " " + str(num_cores) + " " + RCD_location + " -x dunbrack.bin --loco loco.score -o RCD -d " + str(RCD_dist_tol) + " -n " + str(num_loops) + " loops.txt"], shell=True)
            call([RCD_location + " -e 1 -x dunbrack.bin --energy_file loco.score -o RCD -d " + str(RCD_dist_tol) + " -n " + str(num_loops) + " loops.txt  >> rcd.log 2>&1"], shell=True)

            # print("Organizing RCD results")

            call(["mkdir models; cp RCD/anchored_pMHC_closed.pdb models/"], shell=True)        
            os.chdir("models")
            
            call([vina_location + " --input anchored_pMHC_closed.pdb --ligand partial >> vina.log 2>&1"], shell=True)

            array_splits = np.array_split(list(range(1, num_loops+1)), num_cores)
            folder_names = [str(s[0]) for s in array_splits]

            threads = []
            for loop_indices in array_splits:
                t = RefineThread(loop_indices, len(peptide_sequence), num_loops, doReceptorMinimization, flexible_residues, min_with_smina, scoring, debug)
                threads.append(t)
                t.start()
            
            #progress = int(check_output(["for i in " + " ".join(folder_names) + "; do grep \"MODEL\" $i/models_minimize.pdb | wc -l; done | paste -sd+ | bc"], shell=True))
            #print(progress)
            progress = 0
            while progress < num_loops:
                progress = 0
                for s in folder_names:
                    if os.path.exists(s + "/models_minimize.pdb"):
                        progress += int(check_output(["grep \"MODEL\" " + s + "/models_minimize.pdb | wc -l"], shell=True))
                #progress = int(check_output(["for i in " + " ".join(folder_names) + "; do grep \"MODEL\" $i/models_minimize.pdb | wc -l; done | paste -sd+ | bc"], shell=True))
                if progress == 0: continue
                if printProgress: printProgressBar(progress, 100, prefix = 'Progress:', suffix = 'Complete', length = 50)

            #for t in threads: t.join()

            call(["touch models_minimize.pdb"], shell=True)
            if doReceptorMinimization: call(["touch receptor_models_minimize.pdb"], shell=True)
            for folder_name in folder_names:
                call(["cat " + folder_name + "/models_minimize.pdb >> models_minimize.pdb"], shell=True)
                if doReceptorMinimization: call(["cat " + folder_name + "/receptor_new.pdb >> receptor_models_minimize.pdb"], shell=True)


            # find the minimum energy binding mode and use as reference
            f = open("models_minimize.pdb", "r")
            min_energy = 0.0
            model_index = 0
            hasHETATM = False
            min_model_index = -1
            for line in f:
                first_word = line.split()[0]
                if first_word == 'MODEL':
                    f_temp = open("temp.pdb", 'w')
                elif first_word == 'HETATM':
                    hasHETATM = True
                elif first_word == 'ATOM':
                    f_temp.write(line)
                elif first_word == 'ENDMDL':
                    model_index += 1
                    if debug: print("-------------------", model_index)
                    f_temp.close()
                    if hasHETATM:
                        hasHETATM = False
                        continue
                    else:
                        call(["sort -k6 -n temp.pdb > temp_sorted.pdb"], shell=True)
                        if energy < min_energy:
                            try:
                                potential_ref = md.load("temp_sorted.pdb")
                                resi_in_potential_ref = [r for r in potential_ref.top.residues]
                                if len(resi_in_potential_ref) != len(peptide_sequence): continue
                            except: continue
                            ref = potential_ref
                            min_energy = energy
                            min_model_index = model_index
                elif first_word == 'REMARK':
                    if line.split()[1] == 'minimizedAffinity': energy = float(line.split()[2])
            f.close()   

            # print(min_energy)
            if min_energy == 0.0:
                raise ValueError(filename+"\tEnergy filterd failed.")
            ref.save_pdb("../../../confs_top.pdb")
            ref = md.load("../../../confs_top.pdb", atom_indices=ref.top.select("element != H"))
            ref.save_pdb("../../../confs_top.pdb")
            ref_top = ref.top

            if native_loc != None:
                print("Native crystal found!")
                native = get_conf("../../../" + native_loc, ref_top, "element != H", debug)
                print(native)
            else:
                native = None

            call(["cp ../../../confs_top.pdb ../../../peptide.pdb"], shell=True)
            call(["sed -i \"s/          /           /g\" ../../../peptide.pdb"], shell=True) # fix annoying openbabel warning

            process_smina(ref, "../../../conf_data.npz", "../../../confs.dcd", native, min_model_index, debug)      

            os.chdir("..")

            os.chdir("../..")


       
        # print("Loading sampled conformations")
        confs = md.load("confs.dcd", top="confs_top.pdb")
        energies = np.load("conf_data.npz")['energies']
        model_indices = np.load("conf_data.npz")['model_indices']
        min_model_index = np.load("conf_data.npz")['min_model_index']
        
        # print("Num full confs:", len(confs))
        all_confs = confs
        all_energies = energies

        # print("Saving filtered peptide confs")
        if os.path.exists("peptide_confs.pdb"):
            print("Found peptide_confs.pdb, Please move to recompute.")
        else:
            reference_bound = md.load("aln-" + peptide_template)
            ref_backbone = reference_bound.top.select("chainid == 2 and name == 'CA'")
            reference_bound = md.load("aln-" + peptide_template, atom_indices=ref_backbone)
            ref_a1 = reference_bound.xyz[0, 0, :]
            ref_a2 = reference_bound.xyz[0, -1, :]
            ref_a3 = reference_bound.xyz[0, 1, :]
            ref_a4 = reference_bound.xyz[0, -2, :]

            mdtraj_confs = []
            isFirst = True

            filtered_energies = []
            filtered_indices = []
            for i, conf in enumerate(all_confs):
                #print i
                conf.save_pdb("temp.pdb")
                sampled_conf = md.load("temp.pdb")
                sampled_backbone = sampled_conf.top.select("name == 'CA'")
                sampled_conf = md.load("temp.pdb", atom_indices=sampled_backbone)

                a1 = sampled_conf.xyz[0, 0, :]
                a2 = sampled_conf.xyz[0, -1, :]
                a3 = sampled_conf.xyz[0, 1, :]
                a4 = sampled_conf.xyz[0, -2, :]
                d_a1 = np.linalg.norm(a1-ref_a1)*10
                d_a2 = np.linalg.norm(a2-ref_a2)*10
                d_a3 = np.linalg.norm(a3-ref_a3)*10
                d_a4 = np.linalg.norm(a4-ref_a4)*10
                #print i, d_a1, d_a2, energies[i]

                if d_a1 < anchor_tol and d_a2 < anchor_tol and d_a3 < anchor_tol and d_a4 < anchor_tol:
                    #print(i)
                    filtered_energies.append(energies[i])
                    filtered_indices.append(i) # indices used for naming full system confs

                    if isFirst:
                        isFirst = False
                        mdtraj_confs = md.load("temp.pdb")
                    else:
                        mdtraj_confs += md.load("temp.pdb")
            if isinstance(mdtraj_confs,list) and len(mdtraj_confs)==0:
                raise ValueError(filename+"\tAnchor tol filterd failed.")
                
            # print(filtered_energies) 
            # print("Average filtered energy:", np.mean(filtered_energies))
            # print(mdtraj_confs)
            mdtraj_confs.save_pdb("peptide_confs.pdb")
            np.savez_compressed("filtered_energies.npz", filtered_energies=filtered_energies, filtered_indices=filtered_indices)


        filtered_indices = np.load("filtered_energies.npz")["filtered_indices"]
        filtered_energies = np.load("filtered_energies.npz")["filtered_energies"]
        # print("Num filtered confs:", len(filtered_indices))
        # print("Average filtered energy:", np.mean(filtered_energies))

        if saveFullConfs:
            # print("Saving complete peptide-HLA complexes")
            if os.path.exists("full_system_confs"):
                print("Found full_system_confs/ folder. Please move to recompute.")
            else:
                if not doReceptorMinimization:
                    call(["mkdir full_system_confs"], shell=True)
                    for i, conf in enumerate(all_confs):
                        if i not in filtered_indices: continue
                        # print(i)
                        conf.save_pdb("temp.pdb")
                        call(["sed -i \"s/ A  / C  /g\" temp.pdb"], shell=True)
                        call(["cat receptor.pdb temp.pdb | sed \"/MODEL/d\" | sed \"/ENDMDL/d\" | sed \"/END/d\" > system.pdb"], shell=True)
                        call(["cp system.pdb target.pdb"], shell=True)
                        #call(["echo \"14\r\n1\r\" | " + gromacs_location + " -f system.pdb -ignh -o target.pdb > /dev/null 2>&1"], shell=True)
                        call(["cp target.pdb full_system_confs/" + str(i) + ".pdb"], shell=True)
                        #if model_indices[i] == min_model_index: call(["cp target.pdb min_energy_system.pdb"], shell=True)
                    

                else:
                    """ # just use original receptor template as receptor conformation
                    call(["mkdir full_system_confs"], shell=True)
                    for i, conf in enumerate(all_confs):
                        print i
                        conf.save_pdb("temp.pdb")
                        call(["sed -i \"s/ A  / C  /g\" temp.pdb"], shell=True)
                        call(["cat receptor.pdb temp.pdb | sed \"/MODEL/d\" | sed \"/ENDMDL/d\" | sed \"/END/d\" > system.pdb"], shell=True)
                        call(["cp system.pdb target.pdb"], shell=True)
                        #call(["echo \"14\r\n1\r\" | " + gromacs_location + " -f system.pdb -ignh -o target.pdb > /dev/null 2>&1"], shell=True)
                        call(["cp target.pdb full_system_confs/" + str(i) + ".pdb"], shell=True)
                        if model_indices[i] == min_model_index: call(["cp target.pdb min_energy_system.pdb"], shell=True)
                    """
                    
                    os.chdir("RCD/input/models")
                    call([vina_location + " --input receptor_models_minimize.pdb --ligand receptor  >> vina.log 2>&1"], shell=True)
                    os.chdir("../../..")
                    call(["mkdir full_system_confs"], shell=True)

                    for j, conf in enumerate(all_confs):
                        #if printProgress: printProgressBar(j+1, len(all_confs), prefix = 'Progress:', suffix = 'Complete', length = 50)
                        
                        if j not in filtered_indices: continue

                        peptide_j = "temp" + str(model_indices[j]).zfill( len(str(num_loops)) ) + ".pdb"
                        conf.save_pdb(peptide_j)
                        call(["sed -i \"s/ A  / C  /g\" " + peptide_j], shell=True)


                    array_splits = np.array_split(list(range(len(all_confs))), num_cores)
                    folder_names = [str(s[0]) for s in array_splits]
                    num_confs = len(all_confs)

                    threads = []
                    for loop_indices in array_splits:
                        t = ReceptorThread(loop_indices, filtered_indices, model_indices, num_loops)
                        threads.append(t)
                        t.start()
                    for t in threads: t.join()

                    """
                    for j in range(num_confs):
                        
                        if j not in filtered_indices: continue                        

                        receptor_j = "receptor" + str(model_indices[j]).zfill( len(str(num_loops)) ) + ".pdbqt"
                        call(["cp RCD/input/models/" + receptor_j + " ."], shell=True)
                        
                        call(["python " + defaults_location + "/rename_atoms.py " + receptor_j], shell=True) # get receptor_j.temp and receptor_j.complete

                        complex_j = "target" + str(model_indices[j]).zfill( len(str(num_loops)) ) + ".pdb"
                        call(["cat " + receptor_j + ".complete " + peptide_j + " | sed \"/MODEL/d\" | sed \"/ENDMDL/d\" | sed \"/END/d\" > " + complex_j], shell=True)
                        call(["cp " + complex_j + " full_system_confs/" + str(j) + ".pdb"], shell=True)
                        call(["rm " + peptide_j + " " + receptor_j + " " + complex_j + " " + receptor_j + ".temp " + receptor_j + ".complete"], shell=True)
                    """


        # this comes last because calling md.rmsd centers the coordinates (messing up the alignment)
        all_confs = md.load("peptide_confs.pdb")
        all_energies = np.load("filtered_energies.npz")["filtered_energies"]
        alpha_carbon_atoms = all_confs.top.select("name == 'CA'")
        if native != None:
            rmsd_to_native = md.rmsd(all_confs, native, 0) * 10
            print("Min RMSD to native: ", np.min(rmsd_to_native), filtered_indices[np.argmin(rmsd_to_native)])
            if saveFullConfs: call(["cp full_system_confs/" + str(filtered_indices[np.argmin(rmsd_to_native)]) + ".pdb ./minRMSD.pdb"], shell=True)
            print("Energy of MinRMSD: ", all_energies[np.argmin(rmsd_to_native)])
            print("selected binding mode to native: ", md.rmsd(ref, native, 0)[0] * 10)
            rmsds = md.rmsd(all_confs, native, 0) * 10
            rmsds_sort = rmsds[np.argsort(all_energies)]
            pk_unnorm = np.exp(-np.arange(1, len(rmsds)+1))
            pk = pk_unnorm / pk_unnorm.sum()
            print("E-RMSD: ", np.dot(rmsds_sort, pk)) # try computing for just the filtered confs?

            carmsd_to_native = md.rmsd(all_confs, native, 0, atom_indices=alpha_carbon_atoms, ref_atom_indices=alpha_carbon_atoms) * 10
            print("alpha_carbon_MinRMSD: ", np.min(carmsd_to_native), filtered_indices[np.argmin(carmsd_to_native)])
            if saveFullConfs: call(["cp full_system_confs/" + str(filtered_indices[np.argmin(carmsd_to_native)]) + ".pdb ./minCaRMSD.pdb"], shell=True)
            print("Energy of alpha_carbon_MinRMSD: ", all_energies[np.argmin(carmsd_to_native)])
            print("selected binding mode to native alpha_carbon_rmsd: ", md.rmsd(ref, native, 0, atom_indices=alpha_carbon_atoms, ref_atom_indices=alpha_carbon_atoms)[0] * 10)
            rmsds = md.rmsd(all_confs, native, 0, atom_indices=alpha_carbon_atoms, ref_atom_indices=alpha_carbon_atoms) * 10
            rmsds_sort = rmsds[np.argsort(all_energies)]
            pk_unnorm = np.exp(-np.arange(1, len(rmsds)+1))
            pk = pk_unnorm / pk_unnorm.sum()
            print("E-alpha_carbon_RMSD: ", np.dot(rmsds_sort, pk))
        # print("energy of selected binding mode:", np.min(all_energies), filtered_indices[np.argmin(all_energies)])
        if saveFullConfs: call(["cp full_system_confs/" + str(filtered_indices[np.argmin(all_energies)]) + f".pdb {output_dir.strip()}/{filename.strip()}.pdb"], shell=True)
        call([f"sed -i '1i\{np.min(all_energies)}' {output_dir.strip()}/{filename.strip()}.pdb"],shell=True)
        
        if score_with_openmm:
            begin = time.time()

            print("Scoring/Minimizing with OpenMM ...")
            if os.path.exists("full_system_confs/openmm-minimized"):
                print("Found full_system_confs/openmm-minimized folder. Please move to recompute.")
            else:

                os.chdir("full_system_confs/")

                filenames = [str(i)+".pdb" for i in filtered_indices]

                numTries = 10

                min_filenames = []
                energies = []
                for i in range(1, len(filenames)+1):

                    complex_model = "complex-" + str(i).zfill( len(str(len(filenames))) ) + ".pdb"

                    for j in range(numTries):
                        call(["pdbfixer " + filenames[i-1] + " --output=" + complex_model], shell=True)
                        
                        #call(["python " + defaults_location + "/minimize.py " + complex_model + " min-" + complex_model + " > temp.txt"], shell=True)

                        #f = open("temp.txt", 'r')
                        #for line in f:
                        #    line_arr = line.split()
                        #    if line_arr[0] == "total:": energy = float(line_arr[1]) # should go through here twice, second appearance is kept
                        #f.close()

                        energy = minimizeConf(complex_model, "min-" + complex_model, f'{output_dir}/{filename}_data.txt', f'{output_dir}/{filename}_save.pdb', step_size, save_interval, total_time, device)
                        #print(energy)

                        #print(i, filenames[i-1], j, energy)

                        if printProgress: printProgressBar((i-1)*10 + j, len(filenames)*10, prefix = 'Progress:', suffix = 'Complete', length = 50)

                        if energy < 0: 
                            min_filenames.append("min-" + complex_model)
                            energies.append(energy)
                            break
                        else:
                            if j == (numTries-1):
                                call(["rm " + complex_model + " min-" + complex_model], shell=True)


                if printProgress: printProgressBar(len(filenames)*10, len(filenames)*10, prefix = 'Progress:', suffix = 'Complete', length = 50)
                end = time.time()
                print('Openmm process time: ',end-begin)
                call(["mkdir openmm-minimized"], shell=True)
                call(["rm complex-*.pdb"], shell=True)
                call(["mv min-*.pdb openmm-minimized/"], shell=True)
                call(["cp openmm-minimized/" + min_filenames[np.argmin(energies)] + f" {output_dir}/{filename}_openmm.pdb"], shell=True)

                np.savez_compressed("openmm-minimized/openmm_energies.npz", min_filenames=min_filenames, energies=energies)

                os.chdir("..")
    
        if cleanRCD: call(["rm -r RCD"], shell=True)

        os.chdir("..")

    return #sys.exit(0)

def rescore_with_smina(models, receptor, output_loc, doReceptorMinimization, flexible_residues, useSMINA, scoring):

    call([smina_location + " -q --scoring " + scoring + " --out_flex " + output_loc + "/receptor_new.pdb --ligand " + models + " --receptor " + receptor + " --autobox_ligand " + models + " --autobox_add 4 --local_only --minimize --flexres " + flexible_residues + " --energy_range 100 --out " + output_loc + "/models_minimize.pdb  > smina.log 2>&1"], shell=True)

def process_smina(ref, data_name, confs_name, native, min_model_index, debug):

    ref_top = ref.top

    f = open("models_minimize.pdb", "r")
    energies = []
    model_index = 0
    model_indices = []
    for line in f:
        first_word = line.split()[0]
        if first_word == 'MODEL':
            f_temp = open("temp.pdb", 'w')
        elif first_word == 'ATOM':
            f_temp.write(line)
        elif first_word == 'ENDMDL':
            model_index += 1
            if debug: print("-------------------", model_index)
            f_temp.close()
            call(["sort -k6 -n temp.pdb > temp_sorted.pdb"], shell=True)
            try:
                conf = get_conf("temp_sorted.pdb", ref_top, "element != H", debug)
                ref += conf
                energies.append(energy)
                model_indices.append(model_index)
            except:
                continue
        elif first_word == 'REMARK':
            if line.split()[1] == 'minimizedAffinity': energy = float(line.split()[2])
    f.close()   

    confs = ref[1:]
    ref = ref[0]

    # print(confs)

    np.savez_compressed(data_name, energies=energies, model_indices=model_indices, min_model_index=min_model_index) #, rmsd_to_native=rmsd_to_native)
    confs.save_dcd(confs_name)

  
def get_conf(conf_loc, ref_top, selection, debug):

    ref_atoms = [a for a in ref_top.atoms]

    # load conf with no H
    x_top = md.load(conf_loc).top
    x = md.load(conf_loc, atom_indices=x_top.select(selection))

    if x.top.n_atoms != ref_top.n_atoms and debug: 
        print(x.top.n_atoms)
        print(ref_top.n_atoms)
        raise ValueError('x.top.n_atoms != ref_top.n_atoms')

    x_atoms = [a for a in x.top.atoms]

    # rearrange atom lines to link to reference
    new_xyz = np.zeros((1, x.xyz.shape[1], 3))
    for a, ref_atom in enumerate(ref_atoms):
        for x_index, x_atom in enumerate(x_atoms):
            if str(ref_atom) == str(x_atom):
                new_xyz[0, a, :] = x.xyz[0, x_index, :]
                break

    new_x = md.Trajectory(new_xyz, ref_top)
    
    return new_x  


class ReceptorThread(Thread):

    def __init__(self, loop_indices, filtered_indices, model_indices, num_loops):
        self.loop_indices = loop_indices
        self.filtered_indices = filtered_indices
        self.model_indices = model_indices
        self.num_loops = num_loops
        
        Thread.__init__(self)

    def run(self):
        for j in self.loop_indices:
            if j not in self.filtered_indices: continue 
            # print(j)                       

            peptide_j = "temp" + str(self.model_indices[j]).zfill( len(str(self.num_loops)) ) + ".pdb"
            receptor_j = "receptor" + str(self.model_indices[j]).zfill( len(str(self.num_loops)) ) + ".pdbqt"
            call(["cp RCD/input/models/" + receptor_j + " ."], shell=True)
            
            call(["python " + defaults_location + "/rename_atoms.py " + receptor_j], shell=True) # get receptor_j.temp and receptor_j.complete

            complex_j = "target" + str(self.model_indices[j]).zfill( len(str(self.num_loops)) ) + ".pdb"
            call(["cat " + receptor_j + ".complete " + peptide_j + " | sed \"/MODEL/d\" | sed \"/ENDMDL/d\" | sed \"/END/d\" > " + complex_j], shell=True)
            call(["cp " + complex_j + " full_system_confs/" + str(j) + ".pdb"], shell=True)
            call(["rm " + peptide_j + " " + receptor_j + " " + complex_j + " " + receptor_j + ".temp " + receptor_j + ".complete"], shell=True)

            

class RefineThread(Thread):

    def __init__(self, loop_indices, pep_len, num_loops, doReceptorMinimization, flexible_residues, useSMINA, scoring, debug):
        self.loop_indices = loop_indices
        self.pep_len = pep_len
        self.num_loops = num_loops
        self.doReceptorMinimization = doReceptorMinimization
        self.flexible_residues = flexible_residues
        self.useSMINA = useSMINA
        self.scoring = scoring
        self.debug = debug

        if pep_len < 10: self.last_anchor = "\"C   \"" + str(pep_len)
        else:            self.last_anchor = "\"C  \""  + str(pep_len)

        Thread.__init__(self)

    def run(self):
        for i in self.loop_indices:
            if self.debug: print(i)

            model_name_i = "model-" + str(i).zfill( len(str(self.num_loops)) ) + ".pdb"
            partial_name_i = "partial" + str(i).zfill( len(str(self.num_loops)) ) + ".pdbqt"
            fulltemp_model = "fulltemp-" + str(i).zfill( len(str(self.num_loops)) ) + ".pdb"
            complextemp_model = "complextemp-" + str(i).zfill( len(str(self.num_loops)) ) + ".pdb"
            complextemp_model_withH = "complextempwH-" + str(i).zfill( len(str(self.num_loops)) ) + ".pdb"
            full_model = "full-" + str(i).zfill( len(str(self.num_loops)) ) + ".pdb"

            # fill in anchors
            call(["touch " + model_name_i], shell=True)
            call(["grep \"C   1\" ../../../peptide.pdb > " + model_name_i], shell=True)
            call(["cat " + partial_name_i + " >> " + model_name_i], shell=True)
            call(["grep " + self.last_anchor + " ../../../peptide.pdb >> " + model_name_i], shell=True)

            # fill in sidechains
            call([pdbfixer_location + " " + model_name_i + " --output=" + full_model + " --add-atoms=heavy"], shell=True)

            #call(["rm " + model_name_i + " " + partial_name_i], shell=True)
            call(["rm " + partial_name_i], shell=True)

            """ # calling pdbfixer with peptide backbone in receptor, this ends up producing less valid conformations
            call(["cat ../../../receptor.pdb " + model_name_i + " > " + fulltemp_model], shell=True)
            call([pdbfixer_location + " " + fulltemp_model + " --output=" + complextemp_model + " --add-atoms=heavy"], shell=True)
            # create complexes with hydrogen
            #call([pdbfixer_location + " " + fulltemp_model + " --output=" + complextemp_model_withH], shell=True)
            call(["grep \"[A-Z] C \" " + complextemp_model + " > " + full_model], shell=True)

            call(["rm " + model_name_i + " " + partial_name_i + " " + fulltemp_model + " " + complextemp_model], shell=True)
            """

        folder_name = str(self.loop_indices[0])
        call(["mkdir " + folder_name], shell=True)
        call(["touch " + folder_name + "/all_models.pdb"], shell=True)
        for i in self.loop_indices:
            full_model = "full-" + str(i).zfill( len(str(self.num_loops)) ) + ".pdb"
            call(["echo \"MODEL        " + str(i) + "\" >> " + folder_name + "/all_models.pdb"], shell=True)
            call(["sed '/REMARK/d' " + full_model + " | sed '/TER/d' | sed '/END/d' >> " + folder_name + "/all_models.pdb"], shell=True)
            call(["echo \"ENDMDL\" >> " + folder_name + "/all_models.pdb"], shell=True)

        rescore_with_smina(folder_name + "/all_models.pdb", "../../../receptor.pdb", folder_name, self.doReceptorMinimization, self.flexible_residues, self.useSMINA, self.scoring)

# Print iterations progress
def printProgressBar (iteration, total, prefix = '', suffix = '', decimals = 1, length = 100, fill = '*'):
    """
    Call in a loop to create terminal progress bar
    @params:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
        decimals    - Optional  : positive number of decimals in percent complete (Int)
        length      - Optional  : character length of bar (Int)
        fill        - Optional  : bar fill character (Str)
    """
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + '-' * (length - filledLength)
    print('\r%s |%s| %s%% %s' % (prefix, bar, percent, suffix), end = '\r')
    # Print New Line on Complete
    if iteration == total: 
        print()

def minimizeConf(filename, new_filename, data_filename, save_filename, step_size, save_interval, total_time, device='CPU'):
    print('save_interval',save_interval,'step_size',step_size,'total_time',total_time)
    
    #print("Opening:", filename)

    pdb = PDBFile(filename)
    top = pdb.getTopology()
    positions = np.array(pdb.positions) #pdb.getPositions(asNumpy=True)
    numAtoms = len(positions)

    #print("Number of atoms:", numAtoms)
    #print("Number of residues:", top.getNumResidues())

    positions = np.reshape(positions, (3*numAtoms,1))

    # run file through pdb fixer first

    #forcefield = ForceField('amber99sb.xml', 'tip3p.xml')
    #forcefield = app.ForceField('amber99sbildn.xml', 'amber99_obc.xml')
    forcefield = app.ForceField('amber14-all.xml','amber14/tip3p.xml')
    #forcefield = app.ForceField('amber03.xml', 'amber03_obc.xml')
    #forcefield = app.ForceField('amber10.xml', 'amber10_obc.xml')
    modeller = Modeller(pdb.topology, pdb.positions)
    system = forcefield.createSystem(modeller.topology, nonbondedMethod=CutoffNonPeriodic, constraints=None)

    force_constant = 5000
    force = CustomExternalForce("k*periodicdistance(x, y, z, x0, y0, z0)^2")
    force.addGlobalParameter("k", force_constant)
    force.addPerParticleParameter("x0")
    force.addPerParticleParameter("y0")
    force.addPerParticleParameter("z0")
    protein_particles = md.load(filename).top.select("backbone")

    particle_indices = []
    for protein_particle in protein_particles:
        particle_indices.append(force.addParticle(int(protein_particle), modeller.positions[protein_particle]) )
    system.addForce(force)


    forces = system.getForces()
    i = 0
    for f in forces:
        f.setForceGroup(i)
        i = i + 1


    integrator = LangevinIntegrator(310*kelvin, 1/picosecond, step_size*picoseconds)
    #integrator = VerletIntegrator(0.002*picoseconds)
    platform = Platform.getPlatformByName(device)
    simulation = Simulation(modeller.topology, system, integrator, platform)


    simulation.context.setPositions(modeller.positions)

    #printForces(simulation)

    #print("Minimizing energy ... ")
    simulation.minimizeEnergy()

    #printForces(simulation)
    simulation.reporters.append(app.PDBReporter(save_filename,int(save_interval/step_size)))
    simulation.reporters.append(app.StateDataReporter(data_filename, int(save_interval/step_size), step=True, 
    potentialEnergy=True, temperature=True, speed=True, separator='\t'))

    #print "Equilibrating ..."
    totalSteps=int(total_time/step_size)
    simulation.step(totalSteps)

    #if len(sys.argv) == 2: r = PDBReporter('output.pdb', 1)
    #else: 
    r = PDBReporter(new_filename, 1)
    r.report(simulation, simulation.context.getState(getPositions=True, getEnergy=True))


    return simulation.context.getState(getEnergy=True).getPotentialEnergy() / kilojoule_per_mole

if __name__ == "__main__":
    work_dir = os.getcwd()
    try:
        main(sys.argv[1:])
    except Exception:
        pass
    finally:
        call([f'rm -rf {work_dir}/0'],shell=True)



#!/usr/bin/python

#       Sgourakis Lab
#   Author: Santrupti Nerli
#   Date: January 26, 2017
#   Email: snerli@ucsc.edu
#

'''

GRISHIN class contains all the necessary functionalities required to create
Rosetta specific alignment file.

'''

# import other required libraries
import os
import sys
import subprocess

class GRISHIN:

    # class members
    filename = "" # grishin file name
    target_head = "" # target name
    template_head = "" # template name
    target_sequence = "" # target sequence
    template_sequence = "" # template sequence

    # constructor
    def __init__(self, filename, target_head, template_head, target_sequence, template_sequence):
        self.filename = filename
        self.target_head = target_head
        self.template_head = template_head
        self.target_sequence = target_sequence
        self.template_sequence = template_sequence

    # method to create and write to the grishin file
    # the formatting is very specific to Rosetta
    # See the link: https://www.rosettacommons.org/docs/latest/rosetta_basics/file_types/Grishan-format-alignment
    def write(self, is_new = False):
        if is_new:
            writefile = open(self.get_file_name(), "w")
        else:
            writefile = open(self.get_file_name(), "a")
        writefile.write("## "+self.target_head+" "+self.template_head+"\n"+"#"+"\n")
        writefile.write("scores_from_program: 0\n")
        writefile.write("0 "+self.target_sequence+"\n")
        writefile.write("0 "+self.template_sequence+"\n")
        writefile.write("--\n")
        writefile.close()

    # getter method
    def get_file_name(self):
        return self.filename+".grishin"



import torch
import os
import pytorch_lightning as pl
from torch.optim.lr_scheduler import LambdaLR
from torch.utils.data import DataLoader, TensorDataset, Dataset
from sklearn.model_selection import train_test_split
from pandas import DataFrame
from Bio.PDB import PDBParser
from Bio.PDB.Selection import unfold_entities
from Bio import BiopythonWarning
import warnings
warnings.simplefilter('ignore', BiopythonWarning)
aa_idx = {'A': 0, 'C': 1, 'D': 2, 'E': 3, 'F': 4, 'G': 5, 'H': 6, 'I': 7, 'K': 8, 'L': 9, 'M': 10,
          'N': 11, 'P': 12, 'Q': 13, 'R': 14, 'S': 15, 'T': 16, 'V': 17, 'W': 18, 'Y': 19, 'X': 20}


def get_linear_schedule_with_warmup(optimizer, num_warmup_steps, num_training_steps, last_epoch=-1):
    """
    Create a schedule with a learning rate that decreases linearly from the initial lr set in the optimizer to 0, after
    a warmup period during which it increases linearly from 0 to the initial lr set in the optimizer.

    Args:
        optimizer ([`~torch.optim.Optimizer`]):
            The optimizer for which to schedule the learning rate.
        num_warmup_steps (`int`):
            The number of steps for the warmup phase.
        num_training_steps (`int`):
            The total number of training steps.
        last_epoch (`int`, *optional*, defaults to -1):
            The index of the last epoch when resuming training.

    Return:
        `torch.optim.lr_scheduler.LambdaLR` with the appropriate schedule.
    """

    def lr_lambda(current_step: int):
        if current_step < num_warmup_steps:
            return float(current_step) / float(max(1, num_warmup_steps))
        return max(
            0.0, float(num_training_steps - current_step) /
            float(max(1, num_training_steps - num_warmup_steps))
        )

    return LambdaLR(optimizer, lr_lambda, last_epoch)


class CNN_datamodule(pl.LightningDataModule):
    def __init__(self, config, datas, labels, train_idx, val_idx=None):
        super().__init__()
        self.batch_size = config['batch_size']
        self.config = config
        self.datas = datas
        self.labels = labels
        self.train_idx = train_idx
        self.val_idx = val_idx

    def prepare_data(self):
        data = self.datas
        label = self.labels
        self.data_train = data[self.train_idx]
        self.label_train = label[self.train_idx]
        if self.val_idx is not None:
            self.data_val = data[self.val_idx]
            self.label_val = label[self.val_idx]
        else:
            self.data_train, self.data_val, self.label_train, self.label_val = train_test_split(
                self.data_train, self.label_train, test_size=0.1, random_state=22)
        train_num = self.data_train.shape[0]
        self.data_train = self.data_train.reshape(
            [train_num, 1, -1, self.config['width']])
        val_num = self.data_val.shape[0]
        self.data_val = self.data_val.reshape(
            [val_num, 1, -1, self.config['width']])

        self.train_dataset = TensorDataset(torch.FloatTensor(
            self.data_train), torch.LongTensor(self.label_train))
        self.val_dataset = TensorDataset(torch.FloatTensor(
            self.data_val), torch.LongTensor(self.label_val))

    def setup(self, stage):
        pass

    def train_dataloader(self):
        return DataLoader(self.train_dataset, batch_size=self.batch_size, shuffle=True, num_workers=4)

    def val_dataloader(self):
        return DataLoader(self.val_dataset, batch_size=self.batch_size, shuffle=False, num_workers=4)


class Process():
    def __init__(self, seqs, labels=None, task_type='class', seq_len=49):
        self.seqs = seqs
        self.labels = labels
        self.task_type = task_type
        self.seq_len = seq_len

    @staticmethod
    def pad_seq(seq, seq_len):
        if len(seq) < seq_len:
            padding_len = seq_len - len(seq)
            padding_seq = 'X'*padding_len
            seq = seq + padding_seq
        else:
            seq = seq[:seq_len]
        assert len(seq) == seq_len
        return seq

    def seqs_aa2idx(self):
        seqs_idx = []
        for seq in self.seqs:
            seq = self.pad_seq(seq, seq_len=self.seq_len)
            seqs_idx.append([aa_idx[aa] for aa in seq])
        return torch.LongTensor(seqs_idx)

    def label2tensor(self):
        if self.task_type == 'class':
            y = [int(label) for label in self.labels]
            return torch.LongTensor(y)
        elif self.task_type == 'regression':
            return torch.FloatTensor(self.labels)


class RNN_datamodule(pl.LightningDataModule):
    def __init__(self, config, seqs, labels, train_idx, val_idx=None, task_type='class'):
        super().__init__()
        self.batch_size = config['batch_size']
        self.config = config
        self.seqs = seqs
        self.labels = labels
        self.train_idx = train_idx
        self.val_idx = val_idx
        self.task_type = task_type

    def prepare_data(self):
        seqs = self.seqs
        labels = self.labels

        process = Process(seqs, labels, task_type=self.task_type)
        data = process.seqs_aa2idx()
        label = process.label2tensor()
        self.data_train = data[self.train_idx]
        self.label_train = label[self.train_idx]
        if self.val_idx is not None:
            self.data_val = data[self.val_idx]
            self.label_val = label[self.val_idx]
        else:
            self.data_train, self.data_val, self.label_train, self.label_val = train_test_split(
                self.data_train, self.label_train, test_size=0.1, random_state=22)

        self.train_dataset = TensorDataset(self.data_train, self.label_train)
        self.val_dataset = TensorDataset(self.data_val, self.label_val)

    def setup(self, stage):
        pass

    def train_dataloader(self):
        return DataLoader(self.train_dataset, batch_size=self.batch_size, shuffle=True)

    def val_dataloader(self):
        return DataLoader(self.val_dataset, batch_size=self.batch_size, shuffle=False)


class Fusion_dataset(Dataset):
    def __init__(self, cnn_data, rnn_data, label):
        self.cnn_x = torch.FloatTensor(cnn_data)
        self.rnn_x = torch.tensor(rnn_data)
        self.label = torch.tensor(label)
        assert len(cnn_data) == len(rnn_data)

    def __len__(self):
        return len(self.cnn_x)

    def __getitem__(self, index):
        data = {'cnn_input': self.cnn_x[index],
                'rnn_input': self.rnn_x[index], 'label': self.label[index]}
        return data


class Fusion_datamodule(pl.LightningDataModule):
    def __init__(self, config, cnn_data, rnn_data, labels, train_idx, val_idx=None):
        super().__init__()
        self.config = config
        self.batch_size = config['batch_size']
        self.cnn_data = cnn_data
        self.rnn_data = rnn_data
        self.labels = labels
        self.train_idx = train_idx
        self.val_idx = val_idx

    def prepare_data(self):
        cnn_datamod = CNN_datamodule(
            self.config, self.cnn_data, self.labels, self.train_idx, self.val_idx)
        rnn_datamod = RNN_datamodule(
            self.config, self.rnn_data, self.labels, self.train_idx, self.val_idx)
        cnn_datamod.prepare_data()
        rnn_datamod.prepare_data()
        train_labels = rnn_datamod.label_train
        val_labels = rnn_datamod.label_val
        cnn_train = cnn_datamod.data_train
        cnn_val = cnn_datamod.data_val
        rnn_train = rnn_datamod.data_train
        rnn_val = rnn_datamod.data_val

        self.train_dataset = Fusion_dataset(
            cnn_data=cnn_train, rnn_data=rnn_train, label=train_labels)
        self.val_dataset = Fusion_dataset(
            cnn_data=cnn_val, rnn_data=rnn_val, label=val_labels)

    def train_dataloader(self):
        return DataLoader(self.train_dataset, batch_size=self.batch_size, shuffle=True)

    def val_dataloader(self):
        return DataLoader(self.val_dataset, batch_size=self.batch_size, shuffle=False)


def parse_pdb(file) -> DataFrame:
    code = os.path.split(file)[-1].split('.')[0]
    p = PDBParser(PERMISSIVE=1)
    s = p.get_structure(code, file)
    model = s[0]

    atoms = unfold_entities(model, 'A')
    atomHetfield = []
    atomNum = []
    atomName = []
    # altLoc = []

    chainID = []
    resName = []
    resNum = []
    X = []
    Y = []
    Z = []

    for atom in atoms:
        atom_hetfield = atom.full_id[3][0].strip()
        atomHetfield.append(atom_hetfield)
        atom_num = atom.serial_number
        atomNum.append(atom_num)
        atom_name = atom.name
        atomName.append(atom_name[0])

        chain_id = atom.full_id[2]
        chainID.append(chain_id)
        res = atom.parent
        res_name = res.resname
        resName.append(res_name)
        res_num = res.id[1]
        resNum.append(res_num)
        coord = atom.coord
        X.append(coord[0])
        Y.append(coord[1])
        Z.append(coord[2])

    data = {
        'Atom_num': atomNum,
        'Atom_hetfield': atomHetfield,
        'Atom_name': atomName,
        'chain_ID': chainID,
        'Res_name': resName,
        'Res_num': resNum,
        'X': X,
        'Y': Y,
        'Z': Z,
    }

    return DataFrame(data)


#!/usr/bin/python

#       Sgourakis Lab
#   Author: Santrupti Nerli
#   Date: August 22, 2018
#   Email: snerli@ucsc.edu
#

'''

MHC class contains all the necessary functionalities required to fetch
necessary mhc sequences and their respective headers.

'''

# Load the Rosetta commands for use in the Python shell
from database.HLA_sequences_180 import hla_sequences_180
from database.HLA_sequences_complex import hla_sequences
from input_output.input.header_readers import HEADER_READER

# inherit from HEADER_READER, a generic class that reads
# headers from a file
class MHC(HEADER_READER):

    # class member
    trim_mhc = True # to check if mhc sequence needs to be trimmed

    # constructor
    def __init__(self, filename, trim_mhc):
        super().__init__(filename)
        super().read_headers()
        self.trim_mhc = trim_mhc

    # getter methods
    # based on the trim flag, a full sequencce or a trimmed
    # sequence is returned
    def get_sequence(self, header):
        if self.trim_mhc == True:
            return hla_sequences_180[header]
        else:
            return hla_sequences[header]

    # a user can just specify all in the input mhc list file
    # if it is specified, a full list of headers is returned
    # otherwise, only a requested subset of headers is returned
    def get_headers(self):
        if super().get_headers()[0].lower != "all":
            return super().get_headers()
        else:
            if self.trim_mhc == True:
                return hla_sequences_180.keys()
            else:
                return hla_sequences.keys()



import sys
import os.path as osp
import mdtraj as md
import os
import logging
import argparse
from subprocess import run

tri2single = {'Ala': 'A', 'Arg': 'R', 'Asn': 'N', 'Asp': 'D', 'Cys': 'C', 'Gln': 'Q', 'Glu': 'E', 'Gly': 'G', 'His': 'H', 'Ile': 'I',
              'Leu': 'L', 'Lys': 'K', 'Met': 'M', 'Phe': 'F', 'Pro': 'P', 'Ser': 'S', 'Thr': 'T', 'Trp': 'W', 'Tyr': 'Y', 'Val': 'V'}
DEFAULT_DIR = osp.dirname(osp.realpath(__file__))
WORK_DIR = os.getcwd()
APE_LOCATION = osp.join(DEFAULT_DIR, 'APE-Gen')
ROSETTA_LOCATION = osp.join(DEFAULT_DIR, 'mhc-pep-threader')
template_dic = {}
with open(osp.join(DEFAULT_DIR, 'templates.txt')) as f:
    for line in f:
        ls = line.split()
        template_dic[ls[0]] = ls[1]
with open(osp.join(DEFAULT_DIR, 'supported_alleles.txt')) as f:
    supported_hlas = f.read().splitlines()


parser = argparse.ArgumentParser(
    description="Peptide-MHC complex structure Generator", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('--software', default='APE-Gen', choices=[
                    'APE-Gen', 'RosettaMHC'], help="Software used for pMHC structure generation, defaults to 'APE-Gen'")
parser.add_argument('peptide_input', type=str,
                    nargs=1, help='Sequence of peptide')
parser.add_argument('receptor_class', type=str,
                    nargs=1, help='MHC class I allele')
parser.add_argument('--output_dir', type=str, default='structures',
                    help="Output PDB file saved path")
parser.add_argument('--filename', type=str, default='pmhc001',
                    help="Saved PDB file name,e.g.,default='pmhc001', and the output file will be like 'pmhc001.pdb'")
parser.add_argument('--list_supported_hlas', action='store_true',
                    help="List all the supported 110 HLA alleles")

rosetta_args = parser.add_argument_group(title="RosettaMHC simple arguments")
rosetta_args.add_argument(
    "--template_pdb", help="Provide template structure in PDB to perform threading")
rosetta_args.add_argument("--mhc_chain", type=str, default='A',
                          help="HLA chain in the template PDB file")
rosetta_args.add_argument("--peptide_chain", type=str, default='C',
                          help="Peptide chain in the template PDB file")


def check_hlas(hlas):
    for hla in hlas:
        if hla not in supported_hlas:
            raise (ValueError(
                f"Unrecognized hla {hla} for SETTMHC, use '--list_supported_hlas' for more details"))


def run_APE(peptide, hla, outputdir, filename):
    cmd = f"python {APE_LOCATION}/APE_Gen.py {peptide} {hla} --output_dir {outputdir} --filename {filename} --no_progress"
    run(cmd, shell=True)


def run_RosettaMHC(peptide, hla, outputdir, filename, template_pdb, mhc_chain='A', pep_chain='C'):
    with open('./pep_list', 'w') as f:
        f.write(f'>{filename}' + '\n' + peptide + '\n')
    with open('./mhc_list', 'w') as f:
        f.write(hla.split('HLA-')[-1] + '\n')
    template_name = osp.split(template_pdb)[-1].split('.')[0]
    cmd = f"python {ROSETTA_LOCATION}/main.py -relax_after_threading -template_pdb {template_pdb} -mhcs mhc_list -peptides pep_list \
        -mhc_chain {mhc_chain} -peptide_chain {pep_chain} -pep_start_index 181 -interface_cutpoint 180"
    run(cmd, shell=True)
    hla_header = hla.split('HLA-')[-1]
    rosetta_output = f'{hla_header}_none_{filename}_none_none_on_{template_name}.pdb'
    run(f"cp {rosetta_output} {outputdir}/{filename}.pdb", shell=True)


def check_structure(file, true_pep, is_ape=True):
    is_passed = True
    pdb = md.load_pdb(file)
    top = pdb.topology
    table, _ = top.to_dataframe()
    if is_ape:
        pep_sele = top.select('chainid 2')
    else:
        pep_sele = top.select(f'resid 180 to 200')
    table2 = table.loc[pep_sele]
    pep_table = table2.drop_duplicates('resSeq')
    pep_names = pep_table['resName']
    ls = []
    for aa in pep_names:
        aa_one = tri2single.get(aa.capitalize(), 'X')
        ls.append(aa_one)
    pdb_pep = ''.join(ls)
    if pdb_pep != true_pep:
        is_passed = False
    return is_passed


def structure_generator(argv=sys.argv[1:]):
    args = parser.parse_args(argv)
    if args.list_supported_hlas:
        print(supported_hlas)
        return
    check_hlas(hlas=[args.receptor_class])

    software = args.software
    outputdir = args.output_dir
    outputdir = osp.realpath(outputdir)
    filename = args.filename
    output_file = osp.join(outputdir, f'{filename}.pdb')
    peptide = args.peptide_input[0]
    hla = args.receptor_class[0]
    if not osp.exists(outputdir):
        os.mkdir(outputdir)
    if software == 'APE-Gen':
        ape_dir = 'ape_workdir'
        ape_dir = osp.realpath(ape_dir)
        if not osp.exists(ape_dir):
            os.mkdir(ape_dir)
        os.chdir(ape_dir)
        run_APE(peptide, hla, outputdir, filename)
        is_passed = check_structure(output_file, peptide, is_ape=True)
    elif software == 'RosettaMHC':
        template_pdb = osp.realpath(args.template_pdb)
        rosetta_dir = 'rosetta_workdir'
        rosetta_dir = osp.realpath(rosetta_dir)
        if not osp.exists(rosetta_dir):
            os.mkdir(rosetta_dir)
        os.chdir(rosetta_dir)
        run_RosettaMHC(peptide, hla, outputdir, filename,
                       template_pdb, args.mhc_chain, args.peptide_chain)
        is_passed = check_structure(output_file, peptide, is_ape=False)
    if not is_passed:
        print(f"Check for {peptide}_{hla} complex failed")


def make_structure(peptide, hla, anno, structure_dir=None):
    if structure_dir is None:
        structure_dir = 'structures'
    structure_dir = osp.realpath(structure_dir)
    if not osp.exists(structure_dir):
        os.mkdir(structure_dir)
    rosetta_dir = 'rosetta_workdir'
    rosetta_dir = osp.realpath(rosetta_dir)
    if not osp.exists(rosetta_dir):
        os.mkdir(rosetta_dir)
    ape_dir = 'ape_workdir'
    ape_dir = osp.realpath(ape_dir)
    if not osp.exists(ape_dir):
        os.mkdir(ape_dir)

    is_passed = False
    # run APE-Gen
    output_file = osp.join(structure_dir, f'{anno}.pdb')
    os.chdir(ape_dir)
    run_APE(peptide, hla, structure_dir, anno)
    if osp.exists(output_file):
        is_passed = check_structure(output_file, peptide, is_ape=True)
    if not is_passed or not osp.exists(output_file):
        # run RosettaMHC
        os.chdir(rosetta_dir)
        template_pdb = template_dic[hla]
        template_pdb_file = osp.join(APE_LOCATION, f'templates/{template_pdb}')
        run_RosettaMHC(peptide, hla, structure_dir, anno, template_pdb_file)
        is_passed = check_structure(output_file, peptide, is_ape=False)
        if not is_passed:
            logging.error(
                f"Generation for {peptide}_{hla} pMHC structure failed! This pair will show scores '0' in the results")
            run(f"rm {structure_dir}/{anno}.pdb", shell=True)

    os.chdir(WORK_DIR)
    if is_passed:
        return output_file



from subprocess import call
import time
import sys
import os

from Bio import pairwise2
from Bio import SeqIO
from Bio.PDB import *
import re
import argparse

import bs4
import requests

try:
    from modeller import *
    from modeller.automodel import *
except:
    print("Error with importing Modeller: Make sure license key is correct.")
    sys.exit(0)

defaults_location = os.path.dirname(os.path.abspath(__file__))

# python model_receptor.py < fasta file containing alpha chain> <template pdb>

def main(args):

    parser = argparse.ArgumentParser(description="Homology Modeling of HLAs using Modeller", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('alpha_chain_seq', type=str, nargs=1, help='Fasta file (.fasta) containing the sequence of the alpha chain of HLA or name of HLA allele (ex. HLA-A*02:01). If allele name given, the program will try to download the sequence from EMBL-EBI.')
    parser.add_argument('template', type=str, nargs=1, help='PDB of the template HLA or name of HLA allele (ex. HLA-A*02:01). If allele name given, a template based on the allele\'s supertype (as defined in supertype_templates.csv) will be chosen.')

    parser.add_argument("-n", "--num_models", type=int, default=10, help='Number of models to sample with Modeller')

    args = parser.parse_args(args)

    alpha_chain_seq = args.alpha_chain_seq[0]
    template = args.template[0]

    num_models = args.num_models

    if template[-4:] == ".pdb": 
        print("Using user-inputted template PDB")
        template_pdb = template
    else:
        allele_name = template
        print("Allele to model:", allele_name)
        allele_name = allele_name.replace(":", "")

        supertypes = {}
        representatives = {}
        f = open(defaults_location + "/supertype_templates.csv", 'r')
        for line in f:
            allele, supertype, isRep, pdbid = line.split(",")
            pdbid = pdbid[:4] # remove newline
            if supertype not in supertypes.keys():
                supertypes[supertype] = []
            supertypes[supertype].append(allele)
            if isRep == "representative":
                representatives[supertype] = (allele, pdbid)
        f.close()

        #for k in supertypes.keys():
        #    if k != "Unclassified":
        #        print(k, len(supertypes[k]), representatives[k])

        foundAllele = False
        for s in supertypes.keys():
            if allele_name[4:] in supertypes[s]:
                foundAllele = True
                if s == "Unclassified":
                    print("Allele is Unclassified: Using a template from the same serotype")
                    if allele_name[4] == "A":
                        template_allele, template_pdbcode = representatives["A02"]
                    elif allele_name[4] == "B":
                        template_allele, template_pdbcode = representatives["B07"]
                    elif allele_name[4] == "C":
                        template_allele, template_pdbcode = representatives["C00"]
                    else:
                        print("Error: Unrecognized Serotype")
                        sys.exit(0)
                else:
                    print("Obtaining representative template from within supertype:")
                    template_allele, template_pdbcode = representatives[s]
                    print("Template allele, pdbcode:", template_allele, template_pdbcode)
                break
            elif allele_name[4] == "C":
                foundAllele = True
                print("This is an HLA-C allele: Using representative template")
                template_allele, template_pdbcode = representatives["C00"]
                break
                    
        if not foundAllele:
            print("Warning: Allele cannot be found in the internal database ... Defaulting to 2v2w.pdb")
            template_pdbcode = "2v2w"

        print("Downloading " + template_pdbcode)
        call(["python " + defaults_location + "/get_pMHC_pdb.py " + template_pdbcode], shell=True)
        template_pdb = template_pdbcode + ".pdb"    

    print("Removing signal and transmembrane portions of alpha chain seq")

    if alpha_chain_seq[-6:] != ".fasta": # attempt to download

        print("Attempting to Download sequence of allele", alpha_chain_seq)
        page = requests.get("https://www.ebi.ac.uk/cgi-bin/ipd/imgt/hla/get_allele.cgi?" + alpha_chain_seq[4:])
        soup = bs4.BeautifulSoup(page.content, 'lxml')

        list_of_sequences = soup.findAll('pre')
        if len(list_of_sequences) == 0:
            print("Error: Allele could not be found in database")
            sys.exit(0)

        raw_str = str(soup.findAll('pre')[0])
        raw_str = raw_str.replace("\n", "")
        raw_str = raw_str.replace("<pre>", "")
        raw_seq = raw_str.replace("</pre>", "")

    else: 
        alpha_chain_seq_file = alpha_chain_seq
        raw_seq = str(list(SeqIO.parse(alpha_chain_seq_file, "fasta"))[0].seq)
    
    parser = PDBParser()
    structure = parser.get_structure('asfd', template_pdb)
    ppb = PPBuilder()
    for pp in ppb.build_peptides(structure):
        alpha_chain = str(pp.get_sequence())
        break

    """
    alignments = pairwise2.align.localxx(raw_seq, alpha_chain, penalize_end_gaps=False)
    print(alignments[0])
    #sys.exit(0)

    print(alpha_chain[:10])
    print(alpha_chain[-5:])
    print(raw_seq)
    
    print(raw_seq[24:30])
    sys.exit(0)

    if alpha_chain[0] != "M":
        beg_index = [m.start() for m in re.finditer(alpha_chain[:5], raw_seq)][0]
    else:
        beg_index = [m.start() for m in re.finditer(alpha_chain[1:6], raw_seq)][0]
    """

    beg_index = 24 # The signaling portion in the beginning appear to all be the same length
    end_index = [m.start() for m in re.finditer(alpha_chain[-5:], raw_seq)][-1]

    target_sequence = raw_seq[beg_index:end_index+1]

    print("Length of original alpha chain seq:", len(raw_seq))
    print("Length of processed alpha chain seq:", len(target_sequence))
    print("Length of alpha chain in template:", len(alpha_chain))

    """
    f = open(args[0], 'r')
    f_str = ""
    for line in f:
        f_str += line
    f.close()
    target_sequence = "".join(f_str.split())
    """

    target_sequence += "/"
    # constant beta immunoglobin
    target_sequence += "MIQRTPKIQVYSRHPAENGKSNFLNCYVSGFHPSDIEVDLLKNGERIEKVEHSDLSFSKDWSFYLLYYTEFTPTEKDEYACRVNHVTLSQPKIVKWDRDM*" 
    receptor_template = template_pdb

    #log.verbose()
    #log.minimal()
    log.none()

    print("Preparing receptor template")
    #call(["cp " + receptor_template + " receptor_template.pdb"], shell=True)
    #call(["sed -i \"/[A-Z] C  /d\" receptor_template.pdb"], shell=True)
    #call(["python " + defaults_location + "/mod_getseq.py > log.txt"], shell=True) # creates receptor_template.seq
    mod_getseq(receptor_template)

    print("Preparing target sequence")
    f = open("target_sequence.pir", 'w')
    f.write(">P1;target_sequence\n")
    f.write("sequence::     : :     : :::-1.00:-1.00\n")
    f.write(target_sequence)
    f.close()

    print("Aligning target sequence with receptor template")
    starttime = time.time()
    #call(["python " + defaults_location + "/align2d.py >> log.txt"], shell=True) # creates target_sequence-receptor_template.ali
    align_2d()
    endtime = time.time()
    print("Aligning took " + str(endtime - starttime) + " seconds.")

    print("Creating model")
    starttime = time.time()
    #call(["python -u  " + defaults_location + "/model-single-opt.py 2 >> log.txt"], shell=True)
    #call(["cp $(grep \"Top model\" log.txt | awk '{ print $3 }') best_model.pdb"], shell=True)
    best_model, best_score = model_single_opt(num_models)
    call(["cp " + str(best_model) + " best_model.pdb"], shell=True)
    endtime = time.time()
    print("Homology modelling took " + str(endtime - starttime) + " seconds.")

def model_single_opt(num_models):

    #log.verbose()
    #log.minimal()
    env = environ()

    # Give less weight to all soft-sphere restraints:
    env.schedule_scale = physical.values(default=1.0, soft_sphere=0.7)

    #Considering heteroatoms and waters molecules
    env.io.hetatm = env.io.water = True
    # Directories with input atom files:
    env.io.atom_files_directory = './:../atom_files'


    # Modelling 'sequence' with file.ali
    a = automodel(env, alnfile='target_sequence-receptor_template.ali',
                  knowns='receptor_template',
                  sequence='target_sequence',
                  assess_methods=(assess.DOPE, assess.GA341)
                  )
    # Generating 3 models
    a.starting_model = 1
    a.ending_model = int(num_models)

    # Very thorough Variable Target Function Method (VTFM) optimization:
    a.library_schedule = autosched.slow
    a.max_var_iterations = 300

    # Thorough MD optimization:
    a.md_level = refine.slow
     
    # Repeat the whole cycle 2 times and do not stop unless obj.func. > 1E6
    a.repeat_optimization = 2
    a.max_molpdf = 1e6

    a.make()

    # Get a list of all successfully built models from a.outputs
    ok_models = [x for x in a.outputs if x['failure'] is None]

    # Rank the models by DOPE score
    key = 'DOPE score'
    #print(ok_models)
    ok_models.sort(key=lambda a: a[key])

    # Get top model
    m = ok_models[0]
    print("Top model: %s (DOPE score %.100f)" % (m['name'], m[key]))

    return m['name'], m[key]

def align_2d():

    env = environ()
    aln = alignment(env)
    mdl = model(env, file='receptor_template') #, model_segment=('FIRST:A','LAST:C'))
    aln.append_model(mdl, align_codes='receptor_template', atom_files='receptor_template.pdb')
    aln.append(file='target_sequence.pir', align_codes='target_sequence')
    aln.align2d()
    aln.write(file='target_sequence-receptor_template.ali', alignment_format='PIR')
    #aln.write(file='TvLDH-1bdmA.pap', alignment_format='PAP')

def mod_getseq(receptor_template):

    # Get the sequence of the PDB file, and write to an alignment file
    call(["cp " + receptor_template + " receptor_template.pdb"], shell=True)
    call(["sed -i \"/[A-Z] C  /d\" receptor_template.pdb"], shell=True)
    code = 'receptor_template'
    e = environ()
    m = model(e, file=code)
    aln = alignment(e)
    aln.append_model(m, align_codes=code)
    aln.write(file=code+'.seq')


if __name__ == "__main__":
    main(sys.argv[1:])



from simtk.openmm.app import *
from simtk.openmm import *
from simtk.unit import *
from sys import stdout
import sys
import numpy as np
import time
import mdtraj

# USAGE: python minimize.py <pdb file> <output name (optional)>
# minimize while fixing backbone atoms

def main():

    filename = sys.argv[1]

    print("Opening:", filename)

    pdb = PDBFile(filename)
    top = pdb.getTopology()
    positions = np.array(pdb.positions) #pdb.getPositions(asNumpy=True)
    numAtoms = len(positions)

    print("Number of atoms:", numAtoms)
    print("Number of residues:", top.getNumResidues())

    positions = np.reshape(positions, (3*numAtoms,1))

    # run file through pdb fixer first

    #forcefield = ForceField('amber99sb.xml', 'tip3p.xml')
    forcefield = app.ForceField('amber99sbildn.xml', 'amber99_obc.xml')
    #forcefield = app.ForceField('amber03.xml', 'amber03_obc.xml')
    #forcefield = app.ForceField('amber10.xml', 'amber10_obc.xml')
    modeller = Modeller(pdb.topology, pdb.positions)
    system = forcefield.createSystem(modeller.topology, nonbondedMethod=CutoffNonPeriodic, constraints=None)

    force_constant = 5000
    force = CustomExternalForce("k*periodicdistance(x, y, z, x0, y0, z0)^2")
    force.addGlobalParameter("k", force_constant)
    force.addPerParticleParameter("x0")
    force.addPerParticleParameter("y0")
    force.addPerParticleParameter("z0")
    protein_particles = mdtraj.load(filename).top.select("backbone")

    particle_indices = []
    for protein_particle in protein_particles:
        particle_indices.append(force.addParticle(int(protein_particle), modeller.positions[protein_particle]) )
    system.addForce(force)


    forces = system.getForces()
    i = 0
    for f in forces:
        f.setForceGroup(i)
        i = i + 1


    integrator = LangevinIntegrator(300*kelvin, 1/picosecond, 0.002*picoseconds)
    #integrator = VerletIntegrator(0.002*picoseconds)
    platform = Platform.getPlatformByName('OpenCL')
    simulation = Simulation(modeller.topology, system, integrator, platform)


    simulation.context.setPositions(modeller.positions)

    printForces(simulation)

    print("Minimizing energy ... ")
    simulation.minimizeEnergy()

    printForces(simulation)

    simulation.reporters.append(app.StateDataReporter(stdout, 100, step=True, 
    potentialEnergy=True, temperature=True, progress=False, remainingTime=True, 
    speed=True, totalSteps=250000, separator='\t'))

    #print "Equilibrating ..."
    #simulation.step(250000)

    if len(sys.argv) == 2: r = PDBReporter('output.pdb', 1)
    else: r = PDBReporter(sys.argv[2], 1)
    r.report(simulation, simulation.context.getState(getPositions=True, getEnergy=True))

    printForces(simulation)

def printForces(simulation):

    for i in range(simulation.system.getNumForces()):
        f_name = simulation.system.getForce(i).__class__.__name__
        s0 = simulation.context.getState(getEnergy=True, groups=2**i)
        print(f_name + ": " + str(s0.getPotentialEnergy()))

    print("total:", simulation.context.getState(getEnergy=True).getPotentialEnergy())


if __name__ == "__main__":
    main()






#!/usr/bin/env python3

#       Sgourakis Lab
#   Author: Santrupti Nerli
#   Date: February 2, 2018
#   Email: snerli@ucsc.edu
#

# Load the Rosetta commands for use in the Python shell
from pyrosetta import *

#custom libraries
from model import MODEL
from input_output.input.argparser import ARGPARSE
from database.HLA_sequences_180 import hla_sequences_180
from database.HLA_sequences_complex import hla_sequences

# import other required libraries
import os
import sys
import subprocess

'''

Main is the entry point of RosettaMHC

'''

# method to run the RosettaMHC protocol
def run():
    args = ARGPARSE()

    if args.is_list_mhcs() == True:
        if not args.is_no_trim_mhc_flag_set():
            for key,value in hla_sequences.items():
                print(key)
        else:
            for key,value in hla_sequences_180.items():
                print(key)
    else:
        # Load Rosetta database files
        init(options='')
        modeller = MODEL(args)
        modeller.model_mhc_for_each_peptide_beta2m_tcr_chaperone()


if __name__ == "__main__":
    run()



import numpy as np
import sys
import gudhi
import os
import logging
import argparse
from .utils import parse_pdb
from itertools import combinations
from typing import List
from collections import defaultdict
import warnings
from Bio import BiopythonWarning
warnings.simplefilter('ignore', BiopythonWarning)
warnings.filterwarnings("ignore")

Protein_Atom = ['C', 'N', 'O', 'S']
Ligand_Atom = ['C', 'N', 'O', 'S']
Ligand_Atom_pairs = list(combinations(Ligand_Atom, 2))
aa_list = ['ALA', 'ARG', 'ASN', 'ASP', 'CYS', 'GLU', 'GLN', 'GLY', 'HIS', 'HSE', 'HSD', 'SEC',
           'ILE', 'LEU', 'LYS', 'MET', 'PHE', 'PRO', 'SER', 'THR', 'TRP', 'TYR', 'VAL', 'PYL']
sites = [7, 9, 24, 45, 59, 62, 63, 66, 67, 69, 70, 73, 74, 76, 77, 80, 81,
         84, 95, 97, 99, 114, 116, 118, 143, 147, 150, 152, 156, 158, 159, 163, 167, 171]


def get_barcode(dm, pro_num, lig_num, max_length=15.0):
    zero_bar = []
    if pro_num != 0 and lig_num != 0:
        rips_complex = gudhi.RipsComplex(
            distance_matrix=dm, max_edge_length=max_length)
        simplex_tree = rips_complex.create_simplex_tree(max_dimension=1)
        diag = simplex_tree.persistence()

        for j in range(len(diag)):
            temp = [diag[j][1][0], diag[j][1][1]]
            zero_bar.append(temp)
    zero_bar = np.array(zero_bar)
    return zero_bar


def statistic_of_death(bar, left, right):
    t = bar.shape
    if len(bar) == 0:
        return [0, 0, 0, 0, 0, 0]

    dmax = 0
    dmin = 0
    dave = 0
    dstd = 0
    dsum = 0
    dnum = 0
    value = []
    for i in range(t[0]):
        if (bar[i][1] >= left) & (bar[i][1] <= right):
            value.append(bar[i][1])
    if len(value) == 0:
        return [dmax, dmin, dave, dstd, dsum, dnum]
    else:
        dmax = max(value)
        dmin = min(value)
        dave = np.mean(value)
        dstd = np.std(value)
        dsum = sum(value)
        dnum = len(value)
        return [dmax, dmin, dave, dstd, dsum, dnum]


def statistic_of_persistence(bar, left, right):
    t = bar.shape
    if len(bar) == 0:
        return [0, 0, 0, 0, 0, 0]

    pmax = 0
    pmin = 0
    pave = 0
    pstd = 0
    psum = 0
    pnum = 0
    value = []
    for i in range(t[0]):
        if (bar[i][0] >= left) & (bar[i][1] <= right):
            value.append(bar[i][1]-bar[i][0])
    if len(value) == 0:
        return [pmax, pmin, pave, pstd, psum, pnum]
    else:
        pmax = max(value)
        pmin = min(value)
        pave = np.mean(value)
        pstd = np.std(value)
        psum = sum(value)
        pnum = len(value)
        return [pmax, pmin, pave, pstd, psum, pnum]


def get_static_feature(bar, filtration=15.0, unit=0.5):
    column0 = int(filtration/unit)
    N = 6 * 2  # N is the number of statistic features
    feature_matrix = np.zeros((column0, N))
    for n in range(column0):
        death_feature = statistic_of_death(
            bar, unit * n, unit * (n+1))
        persistence_feature = statistic_of_persistence(
            bar, unit * n, unit * (n+1))
        feature_matrix[n, :6] = death_feature
        feature_matrix[n, 6:12] = persistence_feature
    return feature_matrix


def distance(array1, array2):
    assert len(array1) == 3
    assert len(array2) == 3
    return np.linalg.norm(np.array(array1)-np.array(array2))


def get_distance_matrix(pro_atoms, lig_atoms):
    pro_num = len(pro_atoms)
    lig_num = len(lig_atoms)
    assert isinstance(pro_atoms, List)
    assert isinstance(lig_atoms, List)
    all_atoms = np.array(pro_atoms + lig_atoms)
    N = pro_num + lig_num
    dm = np.zeros((N, N))
    pro_inner_dist = 100
    lig_inner_dist = 100
    if pro_num != 0 and lig_num != 0:
        all_atoms = all_atoms[:, :3]
        assert all_atoms.shape[1] == 3
        for i in range(N):
            for j in range(N):
                if i < pro_num:  # pro atom lines
                    if j < pro_num:
                        dm[i, j] = pro_inner_dist
                    else:
                        atom_dist = distance(all_atoms[i], all_atoms[j])
                        dm[i, j] = atom_dist
                else:  # lig atom lines
                    if j >= pro_num:
                        dm[i, j] = lig_inner_dist
                    else:
                        atom_dist = distance(all_atoms[i], all_atoms[j])
                        dm[i, j] = atom_dist
        for i in range(N):
            dm[i, i] = 0
    else:
        pass
    return dm


def parse_pdbdata(pdb_file, pro_chain, lig_chain, save_barcode=False, cutoff=10, max_length=15, filtration=15.0, unit=0.5):

    save_path = os.getcwd()
    code = os.path.split(pdb_file)[-1].split('.')[0]
    logging.info(f'Start parse the file {pdb_file}...')
    data = parse_pdb(file=pdb_file)
    Protein_chain = pro_chain
    Ligand_chain = lig_chain

    # split data into protein_data and ligand_data
    cond = data.chain_ID.isin([Protein_chain])
    pro_data = data[cond]
    cond = pro_data.Atom_name.isin(Protein_Atom)
    pro_data = pro_data[cond]
    cond = pro_data.Res_name.isin(aa_list)
    pro_data = pro_data[cond]
    cond = (pro_data.Atom_hetfield == '')
    pro_data = pro_data[cond]
    pro_resnum = len(np.unique(pro_data.Res_num))
    # assert pro_resnum == pro_data.Res_num.max()
    cond = pro_data.Res_num.isin(sites)  # binding sites
    pro_data = pro_data[cond]
    pro_resnum = len(np.unique(pro_data.Res_num))
    if pro_resnum < 34:
        logging.warning(f'Not enough interface residues in HLA for {code}!')
        return

    cond = data.chain_ID.isin([Ligand_chain])
    lig_data = data[cond]
    cond = lig_data.Atom_name.isin(Ligand_Atom)
    lig_data = lig_data[cond]
    cond = (lig_data.Atom_hetfield == '')
    lig_data = lig_data[cond]

    protein = defaultdict(list)
    ligand = defaultdict(list)
    for tup in pro_data.itertuples():
        atom = tup.Atom_name
        if atom in Protein_Atom:
            protein[atom].append([tup.X, tup.Y, tup.Z])
    for tup in lig_data.itertuples():
        atom = tup.Atom_name
        if atom in Ligand_Atom:
            ligand[atom].append([tup.X, tup.Y, tup.Z])

    total_features = []
    for pro_atom in Protein_Atom:
        pro_atom_ls = protein[pro_atom]

        for lig_atom in Ligand_Atom:
            lig_atom_ls = ligand[lig_atom]
            # protein atom filtered by distance cutoff
            filtered_pro_atoms = []
            for proatom_coord in pro_atom_ls:
                for ligatom_coord in lig_atom_ls:
                    dist = distance(proatom_coord, ligatom_coord)
                    if dist <= cutoff:
                        filtered_pro_atoms.append(proatom_coord)
                        break

            # pro_num = len(protein[pro_atom])
            pro_num = len(filtered_pro_atoms)
            lig_num = len(lig_atom_ls)
            # if pro_num == 0:
            #     logging.warning(
            #         f'Protein has no {pro_atom} in {pro_atom}_{lig_atom} pair!')
            # if lig_num == 0:
            #     logging.warning(
            #         f'Ligand has no {lig_atom} in {pro_atom}_{lig_atom} pair!!')
            total_num = pro_num + lig_num

            all_atoms = np.zeros([total_num, 3])
            if pro_num:
                all_atoms[:pro_num, :] = np.array(filtered_pro_atoms)
            if lig_num:
                all_atoms[pro_num:, :] = np.array(lig_atom_ls)

            dm = get_distance_matrix(filtered_pro_atoms, lig_atom_ls)
            # atom Euclidean distance feature
            zero_bar = get_barcode(
                dm=dm, pro_num=pro_num, lig_num=lig_num, max_length=max_length)
            feature_matrix = get_static_feature(
                bar=zero_bar, filtration=filtration, unit=unit)
            if save_barcode:
                barfile = os.path.join(
                    save_path, f'{code}_{pro_atom}_{lig_atom}_zerobar.csv')
                np.savetxt(barfile, zero_bar, delimiter=',')

            pair_feature = np.array(feature_matrix).reshape(-1,)
            assert pair_feature.shape[0] == 360
            total_features.append(pair_feature)

    total_features = np.array(total_features)
    assert total_features.shape[0] == 16
    assert total_features.shape[1] == 360

    return total_features


def main(args=sys.argv[1:]):
    parser = argparse.ArgumentParser(
        description='A persistent homology method for pMHC interface structure feature generation', formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        '-i', '--input', help='Input file for structure feature generation', required=True)
    parser.add_argument(
        '-o', '--output', help='Output file for pshm results containing the 16×360 feature matrix')
    parser.add_argument('-l', '--length', default=15,
                        type=float, help='Max length for RipsComplex')
    parser.add_argument('-f', '--filtration', default=15, type=float,
                        help='Max length for filtration process, usually the same as RipsComplex max length')
    parser.add_argument('-u', '--unit', default=0.5, type=float,
                        help='Filtration unit (length for each filtration shell)')
    parser.add_argument('-c', '--cutoff', default=10, type=float,
                        help='HLA and peptide atom distance cutoff for HLA atoms filter')
    parser.add_argument('--protein', default='A',
                        help='HLA chain in the PDB file')
    parser.add_argument('--ligand', default='C',
                        help='Peptide chain in the PDB file')
    args = parser.parse_args()
    max_length = args.length
    filtration = args.filtration
    unit = args.unit
    cutoff = args.cutoff
    pro_chain = args.protein
    lig_chain = args.ligand
    features = parse_pdbdata(pdb_file=args.input, pro_chain=pro_chain, lig_chain=lig_chain,
                             cutoff=cutoff, max_length=max_length, filtration=filtration, unit=unit)

    np.savetxt(args.output, features, delimiter=',')




from pyrosetta import *

import os
import sys
import argparse

from movemap import MOVEMAP

def exists(tag, dir):

    for filename in os.listdir(dir):
        if ".movemap" in filename:
            if tag == filename.split(".movemap")[0]:
                return True

    return False

if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='generate movemap')
    parser.add_argument("-pdb", help="provide pdb name", default=None)
    parser.add_argument("-dir", help="provide directory containing pdbs", default=None)
    parser.add_argument("-pattern", help="provide a pattern", default="low")
    parser.add_argument("-pep_start_index", help="provide peptide start index", type=int)
    parser.add_argument("-pep_length", help="provide peptide length", type=int)
    parser.add_argument("-groove_distance", help="provide groove distance", type=float, default=3.5)
    args = parser.parse_args()

    init()

    if (args.pdb == None and args.dir != None):
        for filename in os.listdir(args.dir):
            if args.pattern in filename and ".movemap" not in filename:
                tag = args.dir+"/"+filename.split(".pdb")[0]
                if not exists(tag, args.dir):
                    movemap = MOVEMAP(tag+".pdb", args.pep_start_index, args.pep_length, args.groove_distance, tag+".movemap")
                    movemap.apply()
                else:
                    print ("Ignored: ", tag)
    elif args.pdb != None:
        tag = args.pdb.split(".pdb")[0]
        movemap = MOVEMAP(tag+".pdb", args.pep_start_index, args.pep_length, args.groove_distance, tag+".movemap")
        movemap.apply()



#!/usr/bin/python

#       Sgourakis Lab
#   Author: Santrupti Nerli
#   Date: February 1, 2018
#   Email: snerli@ucsc.edu
#

'''

INTERFACE class contains all the necessary functionalities required to calculate
interface energies between two chains in a pose.

'''

# Load the Rosetta commands for use in the Python shell
from pyrosetta import *
from pyrosetta.rosetta.protocols.analysis import InterfaceAnalyzerMover

class INTERFACE:

    # class members
    pose = None # pose for which we want to calculte binding energies
    ia = None # InterfaceAnalyzerMover object that provides all the interface related interfaces

    # constructor
    def __init__(self, pose):
        self.ia = None
        self.pose = pose

    # method to apply the mover and populate necessary interface energies
    def analyze(self):
        self.ia = InterfaceAnalyzerMover()
        # we could have used self.ia.apply(self.pose), but the apply() function in turn calls
        # apply_const(), so we just call that directly
        self.ia.apply_const(self.pose)

    # getter method
    def get_dG(self):
        # return the interface score (or binding energy)
        return self.ia.get_interface_dG()



#!/usr/bin/python

#       Sgourakis Lab
#   Author: Santrupti Nerli
#   Date: January 26, 2017
#   Email: snerli@ucsc.edu
#

# import biopython libraries
#
# Hamelryck, T., Manderick, B. (2003) PDB parser and structure class implemented in Python. Bioinformatics 19: 2308-2310
from Bio.PDB import *
#

'''

MyPDB class contains all the necessary functionalities required to
return the pdb( or required chain) given the filename.

'''

class MyPDB:

    # class members
    pdb = None # store the pdb read from PDB parser
    filename = "" # pdb filename

    # constructor
    def __init__(self, filename):
        self.filename = filename

    # getter method
    def get_pdb(self):
        return self.pdb

    # read only chain A of the input pdb
    def read_chain_A(self):
        parser = PDBParser()
        structure = parser.get_structure(self.filename, self.filename)
        for model in structure:
            for chain in model:
                self.pdb = chain
                return
